
from numpy import array
from numpy import float64
from numpy import newaxis
from numpy import zeros
# from numpy import hstack
from numpy import cross
from numpy.linalg import norm
from math import sqrt

from scipy.sparse.linalg import spsolve
from scipy.sparse.linalg import splu
# from scipy.sparse.linalg import lsqr
from scipy.sparse import diags

from compas.numerical import fd_numpy
from compas.numerical import connectivity_matrix

from compas.geometry import subtract_vectors
from compas.geometry import length_vector
from compas.geometry import cross_vectors
from compas.geometry import centroid_points


def equilibrium_fdm(form):
    """ Compute equilibrium of the form diagram using the force density method (FDM) with 'q's stored in the form (All coordinates can change).

    Parameters
    ----------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The FormDiagram with attributes relevant to the analysis (force densities, loads, fixities, etc).

    Returns
    -------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The relaxed form diagram.

    """

    # preprocess

    k_i = form.key_index()
    xyz = form.vertices_attributes(('x', 'y', 'z'))
    loads = form.vertices_attributes(('px', 'py', 'pz'))
    q = form.edges_attribute('q')
    fixed = form.vertices_where({'is_fixed': True})
    fixed = [k_i[k] for k in fixed]
    edges = [(k_i[u], k_i[v]) for u, v in form.edges()]

    # compute equilibrium
    # update the mesh geometry

    xyz, q, f, l, r = fd_numpy(xyz, edges, fixed, q, loads)

    for key, attr in form.vertices(True):
        index = k_i[key]
        attr['x'] = xyz[index, 0]
        attr['y'] = xyz[index, 1]
        attr['z'] = xyz[index, 2]

    return form


def vertical_equilibrium_fdm(form, zmax=None):
    """ Compute equilibrium of the form diagram using the force density method (FDM) and update only the z-coordinates.

    Parameters
    ----------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The FormDiagram.
    zmax : float, optional
        The maximum height of the diagram.
        By default no maximum is set.

    Returns
    -------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The scaled form diagram.

    """

    k_i = form.key_index()
    vcount = len(form.vertex)
    anchors = list(form.anchors())
    fixed = list(form.fixed())
    fixed = set(anchors + fixed)
    fixed = [k_i[key] for key in fixed]
    free = list(set(range(vcount)) - set(fixed))
    edges = [(k_i[u], k_i[v]) for u, v in form.edges_where({'_is_edge': True})]
    xyz = array(form.vertices_attributes('xyz'), dtype=float64)
    p = array(form.vertices_attributes(('px', 'py', 'pz')), dtype=float64)
    q = [form.edge_attribute((u, v), 'q') for u, v in form.edges_where({'_is_edge': True})]
    q = array(q, dtype=float64).reshape((-1, 1))
    C = connectivity_matrix(edges, 'csr')
    Ci = C[:, free]
    Cf = C[:, fixed]
    Cit = Ci.transpose()

    Q = diags([q.ravel()], [0])

    A = Cit.dot(Q).dot(Ci)
    B = Cit.dot(Q).dot(Cf)

    xyz[free, 2] = spsolve(A, p[free, 2] - B.dot(xyz[fixed, 2]))

    if zmax:
        factor = max(xyz[free, 2])/zmax
        q = q * factor
        xyz[free, 2] = xyz[free, 2]/factor

        for index, edge in enumerate(form.edges_where({'_is_edge': True})):
            form.edge_attribute(edge, 'q', q[index].item())

    for key, attr in form.vertices(True):
        index = k_i[key]
        attr['z'] = xyz[index, 2]

    return form


def q_from_qid(q, ind, Edinv, Ei, ph):
    r""" Calculate q's from all qid's.

    Parameters
    ----------
    q : array [mx1]
        Force density vector.
    ind : list [k]
        List with the indices of the k independent edges.
    Edinv : array [nix(m - k)]
        Inverse of equilibrium matrix sliced to the dependent edges.
    Ei : array [nix(m - k)]
        Equilibrium matrix sliced to the independent edges.
    ph : array [nix1]
        Stack of the horizontal loads applied to the free nodes [pxi, pyi].

    Returns
    -------
    q : array [mx1]
        Force densities on all edges.

    Notes
    -------
    Solver of the following equation:
    \mathbf{q}_\mathrm{id} = - mathbf{Ed}_\mathrm{id}(\mathbf{E}_\mathrm{i} * \mathbf{q}_\mathrm{i} - \mathbf{p}_\mathrm{h})

    """

    m = len(q)
    dep = list(set(range(m)) - set(ind))
    q[dep] = Edinv.dot(Ei.dot(q[ind]) - ph.reshape(-1, 1))

    return q


def q_from_variables(qid, B, d):
    r""" Calculate q's from the force parameters (independent edges).

    Parameters
    ----------
    qid : array [kx1]
        Force density vector.
    B : array [m x k]
        The linear map on the force densities.
    d : array [mx1]
        A particular solution of the equilibrium.

    Returns
    -------
    q : array [m x 1]
        Force densities on all edges.

    Notes
    ---------
    Check `Block and Lachauer, 2014 <http://dx.doi.org/10.1016/j.mechrescom.2013.11.010>`_ for more details

    """

    q = B @ qid + d

    return q


def xyz_from_q(q, Pi, Xb, Ci, Cit, Cb, SPLU_D=None):
    """ Calculate coordinates's xyz the q's.

    Parameters
    ----------
    q : array [m x 1]
        Force densities of all edges.
    Pi : array [ni x 3]
        External forces applied to the free vertices.
    Xb : array [nb x 3]
        Position of the fixed vertices.
    Ci : array [m x ni]
        Connectivity matrix on the free vertices
    Cit : array [ni x m]
        Transpose of connectivity matrix on the free vertices
    Cb : array [m x nb]
        Connectivity matrix on the fixed vertices
    SPLU_D : callable, optional
        Sparse LU decomposition of CitQCi to speed up optimisation, by default None

    Returns
    -------
    Xfree : array [ni x 3]
        x, y, z coordinates of the nodes

    """

    CiQCb = Cit @ diags(q.flatten()) @ Cb
    if not SPLU_D:
        CitQCi = Cit @ diags(q.flatten()) @ Ci
        SPLU_D = splu(CitQCi)
    Xfree = SPLU_D.solve(Pi - CiQCb.dot(Xb))

    return Xfree


def weights_from_xyz_dict(xyz, tributary_dict, thk=0.5, density=20.0):
    """Compute the trbutary weights based on the geometry of the structure and a tributary dictionary.

    Parameters
    ----------
    xyz : array [n x 3]
        The XYZ coordinates of the thrust network
    tributary_dict : dict
        The dictionary with the list of the triangles to compute the tributary area
    thk : float, optional
        The thickness of the structure to compute the tributary volumes, by default 0.5
    density : float, optional
        The density of the structure to compute the applied vertical loads, by default 20.0

    Returns
    -------
    pz : array [n x 1]
        The vertical loads applied ar each node
    """

    from numpy import zeros

    pz = zeros((len(xyz), 1))

    for i in tributary_dict:
        p0 = xyz[i]
        area = 0
        for j in tributary_dict[i]:
            pj = xyz[j]
            for face in tributary_dict[i][j]:
                face_coord = [xyz[pt] for pt in face]
                pf = centroid_points(face_coord)
                v1 = subtract_vectors(pj, p0)
                v2 = subtract_vectors(pf, p0)
                area += length_vector(cross_vectors(v1, v2))
        pz[i] = 0.25 * area * thk * density

    return pz


def weights_from_xyz(xyz, F, V0, V1, V2, thk=0.5, density=20.0):
    """Compute the tributary weights based on the assembled sparse matrices linking the topology

    Parameters
    ----------
    xyz : array [n x 3]
        The XYZ coordinates of the thrust network
    F : array [f x n]
        Linear transformation from ``X`` [n x 3] to ``c`` [f x 3] with the position of the centroids
    V0 : array [g x n]
        Mark the influence of the original point in the calculation
    V1 : array [g x n]
        Mark the influence of the neighbor points the calculation
    V2 : array [g x f]
        Mark the influence of the centroid points the calculation
    thk : float, optional
        The thickness of the structure to compute the tributary volumes, by default 0.5
    density : float, optional
        The density of the structure to compute the applied vertical loads, by default 20.0

    Returns
    -------
    pz : array [n x 1]
        The vertical loads applied ar each node
    """

    v0 = V0.dot(xyz)
    v1 = V1.dot(xyz) - v0
    v2 = V2.dot(F).dot(xyz) - v0

    ag = norm(cross(v1, v2), axis=1)

    pz = 0.25 * thk * density * V0.transpose().dot(ag)

    return pz


def compute_reactions(form, plot=False):
    """ Compute and plot the reaction on the supports.

    Parameters
    ----------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        FormDiagram to calculate the reactions.
    plot : bool, optional
        Either or not plot the reactions, the default is True.


    Returns
    -------
    None
        Form Diagram is modified in place.

    """

    # Mapping

    k_i = form.key_index()
    i_k = form.index_key()

    # Vertices and edges

    n = form.number_of_vertices()
    fixed = [k_i[key] for key in form.fixed()]
    # rol = [k_i[key] for key in form.vertices_where({'is_roller': True})]
    edges = [(k_i[u], k_i[v]) for u, v in form.edges_where({'_is_edge': True})]

    # Co-ordinates and loads

    xyz = zeros((n, 3))
    px = zeros((n, 1))
    py = zeros((n, 1))
    pz = zeros((n, 1))

    for key, vertex in form.vertex.items():
        i = k_i[key]
        xyz[i, :] = form.vertex_coordinates(key)
        px[i] = vertex.get('px', 0)
        py[i] = vertex.get('py', 0)
        pz[i] = vertex.get('pz', 0)

    # C and E matrices

    C = connectivity_matrix(edges, 'csr')
    uvw = C.dot(xyz)
    U = uvw[:, 0]
    V = uvw[:, 1]
    W = uvw[:, 2]
    q = array([form.edge_attribute((u, v), 'q') for u, v in form.edges_where({'_is_edge': True})])[:, newaxis]

    # Horizontal checks

    Rx = C.transpose().dot(U * q.ravel()) - px.ravel()
    Ry = C.transpose().dot(V * q.ravel()) - py.ravel()
    Rz = C.transpose().dot(W * q.ravel()) - pz.ravel()

    eq_node = {key: [round(Rx[k_i[key]], 1), round(Ry[k_i[key]], 1)] for key in form.vertices_where({'is_fixed': True})}

    for i in fixed:
        key = i_k[i]
        form.vertex_attribute(key, '_rx', value=Rx[i])
        form.vertex_attribute(key, '_ry', value=Ry[i])
        form.vertex_attribute(key, '_rz', value=Rz[i])
        pz = form.vertex_attribute(key, 'pz')
        if plot:
            print('Reactions in key: {0} are:'.format(key))
            print(Rx[i], Ry[i], Rz[i])

    for key in form.vertices_where({'rol_x': True}):
        i = k_i[key]
        eq_node[key] = round(Rx[i], 2)
        form.vertex_attribute(key, '_rx', value=Rx[i])
        if plot:
            print('Reactions in Partial X-Key: {0} :'.format(key))
            print(Rx[i])

    for key in form.vertices_where({'rol_y': True}):
        i = k_i[key]
        eq_node[key] = round(Ry[i], 2)
        form.vertex_attribute(key, '_ry', value=Ry[i])
        if plot:
            print('Reactions in Partial Y-Key: {0} :'.format(key))
            print(Ry[i])

    return


def equilibrium_residual(q, M):
    """Computes equilibrium residual

    Parameters
    ----------
    q : array [m x 1]
        Force densities
    M : :class:`~compas_tno.problems.Problem`
        Problem class with matrices

    Returns
    -------
    Rmax
        Maximum nodal residual
    """

    res_x = M.Citx.dot(M.U * q.ravel()) - M.ph[:len(M.free_x)].ravel()
    res_y = M.City.dot(M.V * q.ravel()) - M.ph[:len(M.free_y)].ravel()
    Rmax = max(sqrt(max(res_x**2)), sqrt(max(res_y**2)))
    Rsum = sum(res_x**2 + res_y**2)

    print('residual', Rmax, Rsum)

    return Rmax


def xyz_from_xopt(variables, M):
    """Compute the nodal position and relevant parameters and from the variables of the optimisation and the class of matrices (M)

    Parameters
    ----------
    variables : array
        The ``n`` variables of the optimisation process
    M : :class:`~compas_tno.problems.Problem`
        The relevant matrices to be stored

    Returns
    -------
    M
        The updated relevant matrices
    """

    if isinstance(M, list):
        M = M[0]

    # variables
    k = M.k  # number of force variables
    n = M.n  # number of vertices
    nb = len(M.fixed)  # number of fixed vertices
    # t = M.shape.datashape['t']

    qid = variables[:k]
    check = k

    if 'xyb' in M.variables:
        xyb = variables[check:check + 2*nb]
        check = check + 2*nb
        M.X[M.fixed, :2] = xyb.reshape(-1, 2, order='F')
        # nbxy = nb
    if 'zb' in M.variables:
        zb = variables[check: check + nb]
        check = check + nb
        M.X[M.fixed, [2]] = zb.flatten()
    if 't' in M.variables or 'n' in M.variables:
        # thk = variables[check: check + 1]
        check = check + 1
    if 'lambdh' in M.variables:
        lambdh = variables[check: check + 1]
        M.P[:, [0]] = lambdh * M.px0
        M.P[:, [1]] = lambdh * M.py0
        M.d = lambdh * M.d0
    if 'tub' in M.variables:
        tub = variables[check: check + n]
        M.tub = tub
        check = check + n
    if 'tlb' in M.variables:
        tlb = variables[check: check + n]
        M.tlb = tlb
        check = check + n
    if 'tub_reac' in M.variables:
        tub_reac = variables[check: check + 2*nb].reshape(-1, 1)
        M.tub_reac = tub_reac
        check = check + 2*nb

    M.q = q_from_variables(qid, M.B, M.d)

    # update geometry
    M.X[M.free] = xyz_from_q(M.q, M.P[M.free], M.X[M.fixed], M.Ci, M.Cit, M.Cb)

    return M



from compas_tno.shapes import Shape
from compas_tno.diagrams import FormDiagram
from compas_tno.viewers import Viewer
from compas_tno.analysis import Analysis


# ----------------------------------------
# 1. Shape geometric definition
# ----------------------------------------
spr_angle = 30.0
L = 10.0
thk = 0.50
xy_span = [[0, L], [0, L]]
vault = Shape.create_crossvault(xy_span=xy_span, thk=thk, spr_angle=30)

# ----------------------------------------
# 2. Form diagram geometric definition
# ----------------------------------------
discretisation = 14
form = FormDiagram.create_fan_form(xy_span=xy_span, discretisation=discretisation)

# --------------------------------------------
# 3. Minimum thurst solution and visualisation
# --------------------------------------------
analysis = Analysis.create_minthrust_analysis(form, vault, printout=True)
analysis.apply_selfweight()
analysis.apply_envelope()
analysis.set_up_optimiser()
analysis.run()

view = Viewer(form)
view.settings['scale.reactions'] = 0.004
view.show_solution()

# --------------------------------------------
# 4. Maximum thurst solution and visualisation
# --------------------------------------------
analysis = Analysis.create_maxthrust_analysis(form, vault, printout=True)
analysis.apply_selfweight()
analysis.apply_envelope()
analysis.set_up_optimiser()
analysis.run()

view = Viewer(form)
view.settings['scale.reactions'] = 0.004
view.show_solution()



from compas.geometry import subtract_vectors
from compas.geometry import length_vector
from compas.geometry import cross_vectors
from compas.geometry import normalize_vector
from compas.geometry import is_point_in_convex_polygon_xy
from compas.geometry import centroid_points
from compas.geometry import normal_polygon
from compas.geometry import norm_vector
from compas.geometry import sum_vectors
from compas.geometry import scale_vector
from compas.geometry import angle_vectors

from compas.datastructures import Mesh

import math


class MeshDos(Mesh):
    """The ``MeshDos`` inherits compas ``Mesh`` object and add some features to create Intrados and Extrados.

    Notes
    -----
    A ``MeshDos`` has the following constructor functions

    *   ``from_library`` : Construct the shape from a dictionary with instructions, the library supports the creation of parametric arches, domes, and vaults.
    *   ``from_rhinomesh`` : Construct Extrados, Intrados and Middle surfaces from RhinoMeshes.
    *   ``from_rhinosurface`` : Construct Extrados, Intrados and Middle surfaces using the U and V isolines.
    *   ``from_pointcloud`` : Construct Extrados, Intrados and Middle surfaces interoplating a pointcloud.

    """

    def __init__(self):  # add 'is_outside': False as default
        super(Mesh, self).__init__()

    @classmethod
    def from_mesh(cls, mesh):
        """ Construct object from an existing COMPAS mesh.

        Parameters
        ----------
        mesh : Mesh
            Base mesh.

        Returns
        -------
        mesh
            A mesh MeshDos mesh.
        """

        vertices, faces = mesh.to_vertices_and_faces()

        return cls().from_vertices_and_faces(vertices, faces)

    @classmethod
    def from_formdiagram_attribute(cls, formdiagram, attribute='lb'):
        """Create a MeshDos object from a base form diagram and an attribute.

        Parameters
        ----------
        formdiagram : :class:`~compas_tno.diagrams.FormDiagram`
            The form diagram to copy
        attribute : str, optional
            The attribute to carry to the ``z``, by default 'lb'

        Returns
        -------
        mesh
            MeshDos
        """

        vertices, faces = formdiagram.to_vertices_and_faces()
        mesh = cls().from_vertices_and_faces(vertices, faces)

        for key in formdiagram.vertices():
            z_ = formdiagram.vertex_attribute(key, attribute)
            mesh.vertex_attribute(key, 'z', z_)

        return mesh

    def offset_mesh(self, n=0.1, direction='up', t=0.0):
        """Offset the mesh upwards considering it as intrados.

        Parameters
        ----------
        n : float, optional
            Normal distance of offset, by default 0.1
        direction : str, optional
            Direction of the offset, by default 'up'
        t : float, optional
            Distance considered in the supports, by default 0.0

        Returns
        -------
        mesh
            MeshDos
        """

        offset_list = []
        mesh_copy = self.copy()

        for key in self.vertices():
            normal = self.vertex_attribute(key, 'n')
            # normal = new_mesh.vertex_normal(key)
            z = self.vertex_coordinates(key)[2]
            if self.vertex_attribute(key, 'is_outside'):
                offset_list.append(t)
            else:
                deviation = 1/math.sqrt(1/(1 + (normal[0]**2 + normal[1]**2)/normal[2]**2))
                # print('deviation:', deviation)
                if direction == 'up':
                    offset_list.append(z + n*deviation*norm_vector(normal))  # Experimenting with this normal norm!
                else:
                    offset_list.append(z - n*deviation*norm_vector(normal))  # Experimenting with this normal norm!

        for i, key in enumerate(mesh_copy.vertices()):
            mesh_copy.vertex_attribute(key, 'z', offset_list[i])

        return mesh_copy

    def offset_up_and_down(self, n=1.0, t=0.0):
        """Offset the mesh up and down considering it as middle surface.

        Parameters
        ----------
        n : float, optional
            Percentage of the normal vector to use, by default 1.0.
        t : float, optional
            Distance considered in the supports, by default 0.0

        Returns
        -------
        mesh_ub, mesh_lb
            Extrados mesh and intrados mesh.

        """

        ub_update = []
        lb_update = []
        mesh_ub = self.copy()
        mesh_lb = self.copy()

        for key in self.vertices():
            nub = self.vertex_attribute(key, 'nub')
            nlb = self.vertex_attribute(key, 'nlb')
            z = self.vertex_coordinates(key)[2]
            dev_ub = 1/math.sqrt(1/(1 + (nub[0]**2 + nub[1]**2)/nub[2]**2))
            if self.vertex_attribute(key, 'is_outside'):
                dev_lb = 0.0
            else:
                dev_lb = 1/math.sqrt(1/(1 + (nlb[0]**2 + nlb[1]**2)/nlb[2]**2))
            ub_update.append(z + n * dev_ub * norm_vector(nub))
            lb_update.append(z - n * dev_lb * norm_vector(nlb))

        for i, key in enumerate(mesh_ub.vertices()):
            mesh_ub.vertex_attribute(key, 'z', ub_update[i])
            mesh_lb.vertex_attribute(key, 'z', lb_update[i])

        return mesh_ub, mesh_lb

    def store_normals(self, correct_creases=False, printout=False, plot=False):
        """Store normals in a MeshDos object

        Parameters
        ----------
        correct_creases : bool, optional
            If creases should be corrected, by default False
        printout : bool, optional
            If prints should appear in the screen, by default False
        plot : bool, optional
            If plots with normals, by default False

        Returns
        -------
        None
            Normals added as attributes
        """

        for key in self.vertices():
            self.vertex_attribute(key, 'n', self.vertex_normal(key))

        if correct_creases:
            self.magnify_normal_at_ribs(printout=printout, plot=plot)

        return

    def identify_creases_by_angle(self, deviation=20.0):
        """ Identify creses in the structure based on a threshold angle limit.

        Parameters
        ----------
        deviation : float, optional
            Angle (deg) deviation of normals along an edge so it is considered a crease. Default is 20.0.

        Returns
        -------
        None
            Identification made in the shape
        """

        self.vertices_attribute('is_crease', False)
        self.edges_attribute('is_crease', False)
        for u, v in self.edges():
            faces = self.edge_faces(u, v)
            if faces[0] is not None and faces[1] is not None:
                normals = [self.face_normal(face) for face in faces]
                dev = angle_vectors(normals[0], normals[1], deg=True)
                if dev > deviation:
                    self.vertex_attribute(u, 'is_crease', True)
                    self.vertex_attribute(v, 'is_crease', True)
                    self.edge_attribute((u, v), 'is_crease', True)

        return

    def identify_creases_at_diagonals(self, xy_span=[[0.0, 10.0], [0.0, 10.0]]):
        """Identify creses in the structure based on the diagonal of a rectangular.

        Parameters
        ----------
        xy_span : list, optional
            Span of the pattern, by default [[0.0, 10.0], [0.0, 10.0]]

        Returns
        -------
        None
            Creases are marked with attributes
        """

        self.vertices_attribute('is_crease', False)
        self.edges_attribute('is_crease', False)
        tol = 10E-3

        y1 = xy_span[1][1]
        y0 = xy_span[1][0]
        x1 = xy_span[0][1]
        x0 = xy_span[0][0]

        for key in self.vertices():
            xi, yi, _ = self.vertex_coordinates(key)
            if abs(yi - (y0 + (y1 - y0)/(x1 - x0) * (xi - x0))) <= tol or abs(yi - (y1 - (y1 - y0)/(x1 - x0) * (xi - x0))) <= tol:
                self.vertex_attribute(key, 'is_crease', True)

        for u, v in self.edges():
            if self.vertex_attribute(u, 'is_crease') and self.vertex_attribute(v, 'is_crease'):
                self.edge_attribute((u, v), 'is_crease', True)

        return

    def magnify_normal_at_ribs(self):
        """Magnify normals at ribs.
        """

        keys = list(self.vertices_where({'is_crease': True}))

        for vkey in keys:
            vfaces = self.vertex_faces(vkey, ordered=True)
            vfaces = vfaces + [vfaces[0]]
            right = []
            left = []
            for i in range(len(vfaces) - 2):
                halfedge = self.face_adjacency_halfedge(vfaces[i], vfaces[i + 1])
                if i == 0:
                    right.append(vfaces[i])
                    if self.edge_attribute(halfedge, 'is_crease'):
                        left.append(vfaces[i + 1])
                    else:
                        right.append(vfaces[i + 1])
                else:
                    if self.edge_attribute(halfedge, 'is_crease'):
                        if vfaces[i] in right:
                            left.append(vfaces[i + 1])
                        else:
                            right.append(vfaces[i + 1])
                    else:
                        if vfaces[i] in right:
                            right.append(vfaces[i + 1])
                        else:
                            left.append(vfaces[i + 1])

            normal_right = normalize_vector(centroid_points([self.face_normal(fkey, False) for fkey in right]))
            normal_left = normalize_vector(centroid_points([self.face_normal(fkey, False) for fkey in left]))
            angle = angle_vectors(normal_right, normal_left)/2
            n = scale_vector(normalize_vector(sum_vectors([normal_right, normal_left])), 1/math.cos(angle))
            self.vertex_attribute(vkey, 'n', n)

        return

    def scale_normals_with_ub_lb(self, zub, zlb, tol=10e-3):
        """Scale the normals on the middle surface to match zub and zlb.

        Parameters
        ----------
        zub : array (n x 1)
            Height of the nodes of extrados.
        zlb : array (n x 1)
            Height of the nodes of intrados.
        tol : float, optional
            The tolerance, by default 10e-3

        """

        self.vertices_attribute('is_outside', False)
        i = 0
        for key in self.vertices():
            n = self.vertex_attribute(key, 'n')
            zti = self.vertex_attribute(key, 'z')
            deviation = 1/math.sqrt(1/(1 + (n[0]**2 + n[1]**2)/n[2]**2))
            fac_ub = (zub[i] - zti)/deviation
            fac_lb = (zlb[i] - zti)/deviation
            nub = scale_vector(n, fac_ub/norm_vector(n))
            nlb = scale_vector(n, fac_lb/norm_vector(n))
            self.vertex_attribute(key, 'nub', nub)
            self.vertex_attribute(key, 'nlb', nlb)
            if abs(fac_lb) < tol:
                self.vertex_attribute(key, 'is_outside', True)
            i += 1

        return

    def get_xy_face_normals(self, XY):
        """ Get normal based on the face normal in which the points are in the plan.

        Parameters
        ----------
        XY : list
            List of Points

        Returns
        -------
        normals : list
            List with normals
        """

        face_coords = [self.face_coordinates(fkey) for fkey in self.faces()]
        normals = []
        for xy in XY:
            normals_for_pt = []
            for face_coord in face_coords:
                i = 0
                if is_point_in_convex_polygon_xy(xy, face_coord):
                    normal = normal_polygon(face_coord, unitized=False)
                    normals_for_pt.append(normal)
                    i = i+1
            if len(normals_for_pt) == 1:
                normal = normalize_vector(normals_for_pt[0])
            else:
                normal = normalize_vector(centroid_points(normals_for_pt))
            normals.append(normal)
        return normals

    def round_xy_boundary(self, span=[[0, 10], [0, 10]], tol=10e-4):
        """Prevent rounding errors by rounding boundary vertices

        Parameters
        ----------
        span : list, optional
            Span of the pattern, by default [[0, 10], [0, 10]]
        tol : float, optional
            Tolerance, by default 10e-4

        Returns
        -------
        None
            Pattern modified in place
        """

        for key in self.vertices():
            x, y, z = self.vertex_coordinates(key)
            if abs(x - span[0][0]) < tol:
                self.vertex_attribute(key, 'x', span[0][0])
            if abs(x - span[1][0]) < tol:
                self.vertex_attribute(key, 'x', span[1][0])
            if abs(y - span[0][1]) < tol:
                self.vertex_attribute(key, 'y', span[0][1])
            if abs(y - span[1][1]) < tol:
                self.vertex_attribute(key, 'y', span[1][1])

        return

    def vertex_projected_area(self, key):
        """Compute the projected tributary area of a vertex.

        Parameters
        ----------
        key : int
            The identifier of the vertex.

        Returns
        -------
        float
            The projected tributary area.

        """

        area = 0.

        p0 = self.vertex_coordinates(key)
        p0[2] = 0

        for nbr in self.halfedge[key]:
            p1 = self.vertex_coordinates(nbr)
            p1[2] = 0
            v1 = subtract_vectors(p1, p0)

            fkey = self.halfedge[key][nbr]
            if fkey is not None:
                p2 = self.face_centroid(fkey)
                p2[2] = 0
                v2 = subtract_vectors(p2, p0)
                area += length_vector(cross_vectors(v1, v2))

            fkey = self.halfedge[nbr][key]
            if fkey is not None:
                p3 = self.face_centroid(fkey)
                p3[2] = 0
                v3 = subtract_vectors(p3, p0)
                area += length_vector(cross_vectors(v1, v3))

        return 0.25 * area



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import scriptcontext as sc

import compas_rhino

from compas_tno.diagrams import FormDiagram


__commandname__ = "TNO_form_from_lines"


def RunCommand(is_interactive):

    if 'TNO' not in sc.sticky:
        compas_rhino.display_message('TNO has not been initialised yet.')
        return

    scene = sc.sticky['TNO']['scene']

    guids = compas_rhino.select_lines(message='Select Form Diagram Lines')
    if not guids:
        return

    compas_rhino.rs.HideObjects(guids)

    lines = compas_rhino.get_line_coordinates(guids)
    form = FormDiagram.from_lines(lines)

    scene.purge()
    scene.add(form, name='Form', layer='TNO::FormDiagram')
    scene.update()
    scene.save()


# ==============================================================================
# Main
# ==============================================================================
if __name__ == '__main__':

    RunCommand(True)



from compas.datastructures import Mesh
from compas.geometry import Line
from compas.geometry import Point
from math import radians
from math import sqrt

import json

from compas_tno.shapes import Shape

from compas_view2 import app
from compas_view2.shapes import Arrow
from compas_view2.shapes import Text
from compas_view2.objects import LineObject
from compas_view2.objects import PointObject

from compas.colors import Color


class Viewer(object):
    """A Class for view 3D thrust networks and shapes.

    Parameters
    ----------
    thrust : FormDiagram, optional
        The FormDiagram you want to plot, by default None
    shape : Shape, optional
        The Shape of masonry to plot, by default None

    Attributes
    ----------

    None

    """

    def __init__(self, thrust=None, shape=None, force=None, show_grid=False, **kwargs):

        super().__init__()  # **kwargs
        self.title = 'Viewer'
        self.app = None
        self.thrust = thrust
        self.shape = shape
        self.force = force
        # self.lines = []
        # self.points = []
        # self.vectors = []
        self.settings = {
            'show.thrust': True,
            'show.shape': True,
            'show.reactions': True,
            'show.reactionlabels': True,
            'show.cracks': True,
            'show.vertex.outside': True,

            'camera.target': [5, 5, 0],
            'camera.distance': 35,  # 20,
            'camera.rx': 45,
            'camera.ry': 0,
            'camera.rz': 45,
            'camera.fov': 40,
            'camera.show.grid': show_grid,
            'camera.show.axis': True,

            'size.vertex': 15.0,
            'size.edge.max_thickness': 15.0,
            'size.edge.base_thickness': 1.0,
            'size.edge.normals': 0.5,
            'size.reaction.head_width': 0.1,
            'size.reaction.body_width': 0.01,
            'size.reactionlabel': 20,

            'force.scale': 0.05,
            'force.anchor': [0, -10, 0],

            'scale.reactions': 0.01,
            'scale.loads': 0.01,
            'scale.edge.thk_absolute': 1.0,
            'opacity.shapes': 0.3,  # before 0.5

            'color.edges.thrust': Color.red(),
            'color.edges.reactions': Color.grey().lightened(80),
            'color.vertex.extrados': Color.from_rgb255(0, 125, 0),
            'color.vertex.intrados': Color.from_rgb255(0, 0, 255),
            'color.vertex.outside': Color.from_rgb255(200, 200, 200),
            'color.mesh.intrados': Color.from_rgb255(125, 125, 125),
            'color.mesh.extrados': Color.from_rgb255(125, 125, 125),
            'color.mesh.middle': Color.from_rgb255(125, 125, 125),
            'color.mesh.general': Color.from_rgb255(0, 0, 0),

            'tol.forces': 1e-2,
            'tol.cracks': 1e-3,
        }

    def initiate_app(self, **kwargs):
        """ Initiate the App with the default camera options

        Returns
        -------
        None
            The objects are updated in place
        """

        self.app = app.App(width=1600, height=900, show_grid=self.settings['camera.show.grid'], **kwargs)

        self.app.view.camera.target = self.settings['camera.target']
        self.app.view.camera.distance = self.settings['camera.distance']
        # self.app.view.camera.rx = radians(self.settings['camera.rx'])
        # self.app.view.camera.rz = radians(self.settings['camera.rz'])
        self.app.view.camera.rotation = [radians(self.settings['camera.rx']),
                                         radians(self.settings['camera.ry']),
                                         radians(self.settings['camera.rz'])]
        self.app.view.camera.fov = self.settings['camera.fov']

    def show_solution(self, **kwargs):
        """ Show the thrust network, with the shape according to the settings

        Returns
        -------
        None
            The objects are updated in place
        """

        if not self.app:
            self.initiate_app()

        self.draw_thrust()
        self.draw_cracks()
        self.draw_shape()
        self.draw_reactions()
        self.show()

    def show(self):
        """ Display the viewer in the screen

        Returns
        -------
        None
            The objects are updated in place
        """

        if not self.app:
            raise ValueError('First Initiate the app')

        self.app.show()

    def save(self, path='temp/fig.png', **kwargs):
        """ Save to a path

        Returns
        -------
        None
            The objects are updated in place
        """

        if not self.app:
            raise ValueError('First Initiate the app')

        # self.app.save(path)
        self.app.view.init()
        self.app.record = True
        self.app.view.paint()
        print(self.app.recorded_frames)

    def add(self, item, **kwargs):
        """Add an item to the viewer.

        Parameters
        ----------
        item : Any
            Item to add

        Returns
        -------
        None
            Viewer updated in place

        """

        if not self.app:
            self.initiate_app()

        self.app.add(item, **kwargs)

    def clear(self):
        """Clear the viewer elements

        Returns
        -------
        None
            The objects are updated in place
        """

        if not self.app:
            raise ValueError('First Initiate the app')

        self.app.view.objects = {}

    def scale_edge_thickness(self, factor=1.0):

        self.settings['scale.edge.thk_absolute'] = self.settings['scale.edge.thk_absolute'] * factor

    def draw_form(self, scale_width=True, absolute_scale=True, cull_negative=False, edges=None, **kwargs):
        """Alias to draw thrust network according to the settings

        Parameters
        ----------
        scale_width : bool, optional
            If the lines of the form diagram should be scaled with regards to the force carried, by default True
        absolute_scale : bool, optional
            If the axial forces should be scaled according to an absoulte scale, by default False

        Returns
        -------
        None
            The viewer is updated in place
        """

        self.draw_thrust(scale_width=scale_width, absolute_scale=absolute_scale, cull_negative=cull_negative, edges=edges, **kwargs)

    def draw_thrust(self, scale_width=True, absolute_scale=True, cull_negative=False, edges=None, **kwargs):
        """Draw thrust network according to the settings

        Parameters
        ----------
        scale_width : bool, optional
            If the lines of the form diagram should be scaled with regards to the force carried, by default True
        absolute_scale : bool, optional
            If the axial forces should be scaled according to an absoulte scale, by default False

        Returns
        -------
        None
            The viewer is updated in place
        """

        if not self.app:
            self.initiate_app()
        edges = edges or list(self.thrust.edges_where({'_is_edge': True}))

        base_thick = self.settings['size.edge.base_thickness']
        max_thick = self.settings['size.edge.max_thickness']
        thickness_abs = self.settings['scale.edge.thk_absolute']

        if scale_width:
            forces = [self.thrust.edge_attribute((u, v), 'q') * self.thrust.edge_length(u, v) for u, v in self.thrust.edges_where({'_is_edge': True})]
            fmax = sqrt(max(abs(max(forces)), abs(min(forces))))  # trying sqrt

        thks = []
        forces = []

        for u, v in edges:
            Xu = self.thrust.vertex_coordinates(u)
            Xv = self.thrust.vertex_coordinates(v)
            line = Line(Xu, Xv)
            if cull_negative:
                if min(Xu[2], Xv[2]) < 0:
                    if Xu[2] > Xv[2]:
                        Xa, Xb = Xu, Xv
                    else:
                        Xa, Xb = Xv, Xu
                    coef = - Xb[2]/(Xa[2] - Xb[2])
                    Xi = [Xb[0] + (Xa[0] - Xb[0]) * coef, Xb[1] + (Xa[1] - Xb[1]) * coef, 0.0]
                    line = Line(Xa, Xi)
            if not scale_width:
                self.app.add(line, name=str((u, v)), linewidth=base_thick, linecolor=self.settings['color.edges.thrust'])
                continue
            q = self.thrust.edge_attribute((u, v), 'q')
            length = self.thrust.edge_length(u, v)
            force = sqrt(abs(q*length))
            thk = force/fmax * max_thick
            if absolute_scale:
                thk = force * thickness_abs
            forces.append(force)
            thks.append(thk)
            if force > self.settings['tol.forces'] * 2:
                # print('thk:', thk)
                self.app.add(line, name=str((u, v)), linewidth=thk, linecolor=self.settings['color.edges.thrust'])

        # print('Min / Max thks:', min(thks), max(thks))
        # print('Min / Max forces:', min(forces), max(forces))

    def draw_cracks(self, cull_negative=False, points=None, **kwargs):
        """Draw cracks according to the settings

        Returns
        -------
        None
            The viewer is updated in place.
        """

        if not self.app:
            self.initiate_app()

        intrad = 1
        extrad = 1
        out = 1
        points = points or list(self.thrust.vertices())

        for key in points:
            lb = self.thrust.vertex_attribute(key, 'lb')
            ub = self.thrust.vertex_attribute(key, 'ub')
            x, y, z = self.thrust.vertex_coordinates(key)
            if cull_negative:
                if z < 0.0:
                    continue
            if self.settings['show.cracks']:
                if abs(ub - z) < self.settings['tol.cracks']:
                    self.app.add(Point(x, y, z), name="Extrados (%s)" % extrad, pointcolor=self.settings['color.vertex.extrados'], pointsize=self.settings['size.vertex'])
                    extrad += 1
                elif abs(lb - z) < self.settings['tol.cracks']:
                    self.app.add(Point(x, y, z), name="Intrados (%s)" % intrad, pointcolor=self.settings['color.vertex.intrados'], pointsize=self.settings['size.vertex'])
                    intrad += 1
                if self.thrust.vertex_attribute(key, 'is_fixed'):
                    if z > 0.0:
                        if self.thrust.vertex_attribute(key, 'b'):
                            bx, by = self.thrust.vertex_attribute(key, 'b')
                            rx, ry, rz = self.thrust.vertex_attribute(key, '_rx'), self.thrust.vertex_attribute(key, '_ry'), self.thrust.vertex_attribute(key, '_rz')
                            scale = z/rz
                            if abs(rx) < self.settings['tol.cracks'] and abs(ry) < self.settings['tol.cracks']:
                                continue
                            # print('Reaction', key, rx, ry, rz)
                            crack_Rx = abs(bx) - abs(rx * scale)  # must be >= 0
                            crack_Ry = abs(by) - abs(ry * scale)  # must be >= 0

                            # print('Cracks:', key, crack_Rx, crack_Ry)

                            if crack_Rx > self.settings['tol.cracks'] and crack_Ry > self.settings['tol.cracks']:  # it is not a crack
                                continue
                            crack_point = Point(x - rx * scale, y - ry * scale, 0.0)
                            if abs(abs(bx) - abs(rx * scale)) < self.settings['tol.cracks'] or abs(abs(by) - abs(ry * scale)) < self.settings['tol.cracks']:
                                self.app.add(crack_point, name="Extrados (%s)" % extrad,
                                             pointcolor=self.settings['color.vertex.extrados'], pointsize=self.settings['size.vertex'])
                                extrad += 1
                            else:
                                self.app.add(Point(x, y, z), name="Outside - Intra (%s)" % out, pointcolor=self.settings['color.vertex.outside'],
                                             pointsize=self.settings['size.vertex'])
                                out += 1

            if self.settings['show.vertex.outside']:
                if z > ub:
                    self.app.add(Point(x, y, z), name="Outside - Intra (%s)" % out, pointcolor=self.settings['color.vertex.outside'], pointsize=self.settings['size.vertex'])
                    out += 1
                elif z < lb:
                    self.app.add(Point(x, y, z), name="Outside - Extra (%s)" % out, pointcolor=self.settings['color.vertex.outside'], pointsize=self.settings['size.vertex'])
                    out += 1

    def draw_shape(self, **kwargs):
        """Draw the shape (intrados + extrados) according to the settings

        Returns
        -------
        None
            The Viewer object is modified in place
        """

        if not self.app:
            self.initiate_app()

        if self.shape:
            datashape = self.shape.datashape.copy()
            if datashape['type'] in ['dome']:
                datashape['type'] = 'dome_polar'
                datashape['discretisation'] = [50, 50]
                shape = Shape.from_library(datashape)
                print('Drawing nicer dome')
            elif datashape['type'] == 'arch':
                datashape['type'] = 'arch_polar'
                shape = Shape.from_library(datashape)
                print('Drawing nicer arch')
            else:
                shape = self.shape

        else:
            shape = Shape.from_formdiagram_and_attributes(self.thrust)

        self.app.add(shape.intrados, name="Intrados", show_lines=False, opacity=self.settings['opacity.shapes'], facecolor=self.settings['color.mesh.intrados'])
        self.app.add(shape.extrados, name="Extrados", show_lines=False, opacity=self.settings['opacity.shapes'], facecolor=self.settings['color.mesh.extrados'])

    def draw_middle_shape(self, **kwargs):
        """ Draw the middle of the shape according to the settings

        Returns
        -------
        None
            The Viewer object is modified in place
        """

        if not self.app:
            self.initiate_app()

        shape = self.shape
        if not shape:
            shape = Shape.from_formdiagram_and_attributes(self.thrust)

        vertices_middle, faces_middle = shape.middle.to_vertices_and_faces()
        mesh_middle = Mesh.from_vertices_and_faces(vertices_middle, faces_middle)
        self.app.add(mesh_middle, name="Middle", show_lines=True, opacity=self.settings['opacity.shapes'], facecolor=self.settings['color.mesh.middle'])

    def draw_shape_normals(self, **kwargs):
        """ Draw the shape normals at intrados and extrados surfaces

        Returns
        -------
        None
            The Viewer object is modified in place
        """

        if not self.app:
            self.initiate_app()

        shape = self.shape
        if not shape:
            shape = Shape.from_formdiagram_and_attributes(self.thrust)

        intrados = shape.intrados
        extrados = shape.extrados
        length = self.settings['size.edge.normals']

        for mesh in [intrados, extrados]:
            for key in mesh.vertices():
                n = mesh.vertex_attribute(key, 'n')
                pt0 = mesh.vertex_coordinates(key)
                pt1 = [pt0[0] + n[0]*length, pt0[1] + n[1]*length, pt0[2] + n[2]*length]
                line = Line(pt0, pt1)
                self.app.add(line, name='normal-{}'.format(key))

    def draw_mesh(self, mesh=None, show_edges=True, opacity=0.5, color=Color.grey(), **kwargs):
        """Draw a mesh to the viewer, if no mesh is given the ``self.thrust`` is taken

        Parameters
        ----------
        mesh : Mesh, optional
            Mesh to plot, by default None
        show_edges : bool, optional
            Whether or not edges are shown, by default True
        opacity : float, optional
            The opacity of the mesh, by default 0.5

        Returns
        -------
        None
            The Viewer is updated in place.
        """

        if not self.app:
            self.initiate_app()

        if not mesh:
            mesh = self.thrust

        vertices_mesh, faces_mesh = mesh.to_vertices_and_faces()
        mesh = Mesh.from_vertices_and_faces(vertices_mesh, faces_mesh)

        self.app.add(mesh, show_lines=show_edges, opacity=opacity, facecolor=color)

    def draw_thrustsurface(self, show_edges=False, show_faces=True, opacity=0.2, color=Color.from_rgb255(125, 125, 125), **kwargs):
        """Draw a mesh to the viewer, if no mesh is given the ``self.thrust`` is taken

        Parameters
        ----------
        mesh : Mesh, optional
            Mesh to plot, by default None
        show_faces : bool, optional
            Whether or not faces are shown, by default True
        show_edges : bool, optional
            Whether or not edges are shown, by default False
        opacity : float, optional
            The opacity of the mesh, by default 0.5
        color : Color, optional
            Color of the mesh, by default grey

        Returns
        -------
        None
            The Viewer is updated in place.
        """

        if not self.app:
            self.initiate_app()

        mesh = self.thrust

        vertices_mesh, faces_mesh = mesh.to_vertices_and_faces()
        mesh = Mesh.from_vertices_and_faces(vertices_mesh, faces_mesh)

        self.app.add(mesh, show_lines=show_edges, show_faces=show_faces, opacity=opacity, facecolor=color)

    def draw_reactions(self, emerging_reactions=False, supports=None, extend_reactions=False, scale=None, **kwargs):
        """Draw the reaction vectors on the supports according to the settings

        Returns
        -------
        None
            The viewer is updated in place.
        """

        if not self.app:
            self.initiate_app()

        if self.settings['show.reactions']:
            reaction_scale = scale or self.settings['scale.reactions']
            thickness_scale = self.settings['scale.edge.thk_absolute']

            supports = supports or list(self.thrust.vertices_where({'is_fixed': True}))

            for key in supports:
                x, y, z = self.thrust.vertex_coordinates(key)
                rx = self.thrust.vertex_attribute(key, '_rx') * reaction_scale
                ry = self.thrust.vertex_attribute(key, '_ry') * reaction_scale
                rz = self.thrust.vertex_attribute(key, '_rz') * reaction_scale
                r = 1/reaction_scale * (rx**2 + ry**2 + rz**2)**(1/2)
                if r < 1e-3:
                    continue
                if extend_reactions:
                    if z < 0:
                        continue
                    rs = z / abs(self.thrust.vertex_attribute(key, '_rz'))
                    rx, ry, rz = self.thrust.vertex_attribute(key, '_rx') * rs, self.thrust.vertex_attribute(key, '_ry') * rs, self.thrust.vertex_attribute(key, '_rz') * rs,
                    max_thickness = self.settings['size.edge.max_thickness']
                    max_f = max([abs(self.thrust.edge_attribute(edge, 'q') * self.thrust.edge_length(*edge)) for edge in self.thrust.edges()])
                    thickness = r/max_f*max_thickness

                    thickness = thickness_scale*sqrt(r)

                    line = Line([x, y, z], [x - rx, y - ry, z - rz])
                    self.app.add(line, linewidth=thickness, linecolor=self.settings['color.edges.thrust'])
                    continue
                if emerging_reactions:
                    arrow = Arrow([x, y, z], [-rx, -ry, -rz], head_width=self.settings['size.reaction.head_width'],
                                  body_width=self.settings['size.reaction.body_width'])
                else:
                    arrow = Arrow([x - rx, y - ry, z - rz], [rx, ry, rz], head_width=self.settings['size.reaction.head_width'],
                                  body_width=self.settings['size.reaction.body_width'])
                self.app.add(arrow, linecolor=self.settings['color.edges.reactions'])  # check if this color= works for new compasview2

    def draw_vector(self, vector=None, base=None, **kwargs):
        """Helper to add vectors to the plotter.

        Parameters
        ----------
        vectors : Vector, optional
            Vector to plot, by default None
        bases : Point, optional
            Point with the location of the base of the vector, by default None

        Returns
        -------
        None
            The plotter is updated in place
        """

        if not vector or not base:
            raise ValueError()

        arrow = Arrow(base, vector, head_width=self.settings['size.reaction.head_width'], body_width=self.settings['size.reaction.body_width'])
        self.app.add(arrow, **kwargs)

    def draw_loads(self, **kwargs):
        """Draw the externally applied loadss as vectors on the applied nodes

        Returns
        -------
        None
            The viewer is updated in place.
        """

        if not self.app:
            self.initiate_app()

        if self.settings['show.reactions']:
            reaction_scale = self.settings['scale.loads']

            for key in self.thrust.vertices():
                x, y, z = self.thrust.vertex_coordinates(key)
                px = self.thrust.vertex_attribute(key, 'px') * reaction_scale
                py = self.thrust.vertex_attribute(key, 'py') * reaction_scale
                pz = self.thrust.vertex_attribute(key, 'pz') * reaction_scale
                arrow = Arrow([x - px, y - py, z - pz], [px, py, pz], head_width=self.settings['size.reaction.head_width'], body_width=self.settings['size.reaction.body_width'])
                self.app.add(arrow, linecolor=self.settings['color.edges.reactions'])

    def draw_reaction_label(self, **kwargs):
        """Draw the reaction labels (force magnitude) on the supports according to the settings

        Returns
        -------
        None
            The viewer is updated in place.
        """

        if not self.app:
            self.initiate_app()

        if self.settings['show.reactionlabels']:
            reaction_scale = self.settings['scale.reactions']

            for key in self.thrust.vertices_where({'is_fixed': True}):
                x, y, z = self.thrust.vertex_coordinates(key)
                rx = - self.thrust.vertex_attribute(key, '_rx') * reaction_scale
                ry = - self.thrust.vertex_attribute(key, '_ry') * reaction_scale
                rz = - self.thrust.vertex_attribute(key, '_rz') * reaction_scale
                pt = [x + rx, y + ry, z + rz]
                reaction = '({0:3g}, {1:3g}, {2:3g})'.format(rx, ry, rz)
                text = Text(reaction, pt, height=self.settings['size.reactionlabel'])
                self.app.add(text)

    def draw_force(self, force=None, show_edges=True, opacity=0.5, **kwargs):
        """Add the ForceDiagram to the viewer, if no force is given the force diagram is recomputed in place.

        Parameters
        ----------
        force : ForceDiagram, optional
            ForceDiagram to plot, by default ``None``
        show_edges : bool, optional
            Whether or not edges are shown, by default ``True``
        opacity : float, optional
            The opacity of the mesh, by default 0.5

        Returns
        -------
        None
            The Viewer is updated in place.
        """

        if not self.app:
            self.initiate_app()

        if not force:
            from compas_tno.algorithms import reciprocal_from_form  # here to avoid double imports
            force = reciprocal_from_form(self.thrust)

        from compas.geometry import Translation
        from compas.geometry import Scale

        scale = self.settings['force.scale']
        translation = self.settings['force.anchor']

        S = Scale.from_factors(3 * [scale])
        T = Translation.from_vector(translation)

        force = force.transformed(S)
        force = force.transformed(T)

        self.app.add(force, show_lines=show_edges, opacity=opacity)

    def draw_assembly(self, assembly, **kwargs):
        """Draw the reaction labels (force magnitude) on the supports according to the settings

        Returns
        -------
        None
            The viewer is updated in place.
        """

        if not self.app:
            self.initiate_app()

        for block in assembly.blocks():
            self.add(block, **kwargs)

    def draw_b_constraint(self):

        base_thick = self.settings['size.edge.base_thickness']
        for key in self.thrust.vertices_where({'is_fixed': True}):
            x, y, _ = self.thrust.vertex_coordinates(key)
            b = self.thrust.vertex_attribute(key, 'b')
            if b:
                bx, by = b
                if abs(bx) > 1e-4:
                    line = Line([x, y, 0], [x + bx, y + by, 0])
                    self.app.add(line, linewidth=base_thick)
                if abs(by) > 1e-4:
                    line = Line([x, y, 0], [x + bx, y + by, 0])
                    self.app.add(line)

    def to_objects(self):
        """Generate General objects (dicts) from the added elements in the viewer (WIP).

        Returns
        -------
        data
            Raw geometry objects in a .
        """

        objects = self.app.view.objects

        print(objects)

        data = {}
        lines = []
        points = []

        for obj in objects:
            if isinstance(obj, LineObject):
                lines.append({'start': [obj._data[0].x, obj._data[0].y, obj._data[0].z],
                              'end': [obj._data[1].x, obj._data[1].y, obj._data[1].z],
                              'color': (255*obj.linecolor).tolist(),
                              'width': obj.linewidth
                              })
            if isinstance(obj, PointObject):
                points.append({'pos': [obj._data.x, obj._data.y, obj._data.z],
                               'color': (255*obj.pointcolor).tolist()
                               })

        data['lines'] = lines
        data['points'] = points

        return data

    def to_json(self, file="out.json"):

        data = self.to_objects()

        with open(file, "w") as outfile:
            json.dump(data, outfile)



import time
import json

from compas.geometry import centroid_points
from compas.geometry import subtract_vectors

from compas_tno.plotters import TNOPlotter


def animation_from_optimisation(analysis, show_force=False, settings=None, record=False, interval=100, jump_each=1):
    """Make a 3D animated plot with the optimisation steps.

    Parameters
    ----------
    form : FormDiagram
        Form Diagram in which the animation should be based
    file_Xform : str
        Path for the ``.json`` file with the notal position of the form diagram during iterations
    force : ForceDiagram, optional
        Force diagram in whch the animation should be based, by default None
    file_Xforce : str, optional
        Path for the ``.json`` file with the notal position of the form diagram during iterations, by default None
    shape : Shape, optional
        The shape to plot in the animation, by default None
    settings : dict, optional
        Dictionary with settings to modify the viewer, by default None
    record : bool, optional
        Whether or not the animation should be saved as ``.gif``, by default False
    interval : int, optional
        Time interval (in ms) among iteration frames, by default 100
    jump_each : int, optional
        Interval tot jump among iteration frames, by default 1, in which all frames are shown
    """

    from compas_tno.algorithms import reciprocal_from_form
    from compas_tno.viewers import Viewer
    import compas_tno

    form = analysis.form
    # optimiser = analysis.optimiser
    shape = analysis.shape
    force = None
    if show_force:
        force = reciprocal_from_form(form)
    file_Xform = compas_tno.get('Xform.json')  # analysis.optimiser.Xform
    file_Xforce = compas_tno.get('Xforce.json')  # analysis.optimiser.Xforce

    viewer = Viewer(form, shape=shape)
    if settings:
        viewer.settings = settings
    viewer.initiate_app()

    with open(file_Xform, mode='r', encoding='utf-8') as f:
        Xform = json.load(f)

    if force:
        with open(file_Xforce, mode='r', encoding='utf-8') as f:
            Xforce = json.load(f)

    iterations_total = len(Xform)
    iterations = round(iterations_total / jump_each)

    print('Displaying {0} iterations from a total of {1} iterations'.format(iterations, iterations_total))

    out = None
    if record:
        if isinstance(record, str):
            out = record
        else:
            out = compas_tno.get('out.gif')
        print('Save gif at:', out)

    @viewer.app.on(interval=interval, frames=iterations, record=record, record_path=out)
    def update(f):

        print(f)

        if f == 1:
            time.sleep(5)

        viewer.clear()

        Xf = Xform[str(f * jump_each)]
        index = 0
        for vertex in form.vertices():
            viewer.thrust.vertex_attribute(vertex, 'x', Xf[index][0])
            viewer.thrust.vertex_attribute(vertex, 'y', Xf[index][1])
            viewer.thrust.vertex_attribute(vertex, 'z', Xf[index][2])
            index += 1

        viewer.draw_thrust()
        # viewer.draw_mesh(viewer.thrust)
        # viewer.draw_cracks()
        # viewer.draw_reactions()
        # viewer.draw_loads()
        # viewer.draw_shape()
        # viewer.draw_reactions()

        if show_force:
            _Xf = Xforce[str(f * jump_each)]
            index = 0
            for vertex in force.vertices():
                force.vertex_attribute(vertex, 'x', _Xf[index][0])
                force.vertex_attribute(vertex, 'y', _Xf[index][1])
                force.vertex_attribute(vertex, 'z', 0.0)
                index += 1
            force_centroid = centroid_points(force.vertices_attributes('xyz'))
            force_centroid0 = [-6.991104169236619, 66.12107435370979, 0.0]  # by hand because it isn't kept as a variable (it's deleted @ each passage)
            dc = subtract_vectors(force_centroid, force_centroid0)
            for vertex in force.vertices():
                x, y, _ = force.vertex_coordinates(vertex)
                force.vertex_attribute(vertex, 'x', x - dc[0])
                force.vertex_attribute(vertex, 'y', y - dc[1])
            viewer.draw_force(force)

    viewer.app.run()

    return


def animation_from_section(form, file_Xform):
    """Make a 3D animated plot with the xz section of the solution.

    Parameters
    ----------
    form : FormDiagram
        Form Diagram in which the animation should be based
    file_Xform : str
        Path for the ``.json`` file with the notal position of the form diagram during iterations

    """

    with open(file_Xform, mode='r', encoding='utf-8') as f:
        Xform = json.load(f)

    iterations = len(Xform)
    print('number of iterations', iterations)

    for i in range(iterations):

        print('Iteration:', i)

        Xi = Xform[str(i)]
        index = 0
        for vertex in form.vertices():
            form.vertex_attribute(vertex, 'x', Xi[index][0])
            form.vertex_attribute(vertex, 'y', Xi[index][1])
            form.vertex_attribute(vertex, 'z', Xi[index][2])
            index += 1

        plotter = TNOPlotter()
        meshartist = plotter.add(form)
        meshartist.draw_edges()
        plotter.show()

    return



if __name__ == '__main__':
    import compas_rhino
    from compas_rhino.uninstall import uninstall
    from compas_rhino.uninstall_plugin import uninstall_plugin
    import os

    print("\n", "-"*10, "Removing existing plugins", "-"*10)
    python_plugins_path = compas_rhino._get_python_plugins_path("6.0")
    print("Plugin location: ", python_plugins_path)
    plugins = os.listdir(python_plugins_path)
    for p in plugins:
        uninstall_plugin(p, version="6.0")

    print("\n", "-"*10, "Removing existing packages", "-"*10)
    uninstall()

    print("\n", "-"*10, "Uninstallation is successful", "-"*10)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import scriptcontext as sc

import compas_rhino
from compas_tno.optimisers import Optimiser
from compas_tno.rhino import OptimiserObject
from compas_tno.rhino import SettingsForm


__commandname__ = "TNO_optimization_settings"


def RunCommand(is_interactive):

    if 'TNO' not in sc.sticky:
        compas_rhino.display_message('TNO has not been initialised yet.')
        return

    scene = sc.sticky['TNO']['scene']

    objects = scene.find_by_name('Optimiser')
    if not objects:
        optimiser = Optimiser()
        scene.add(optimiser, name='Optimiser', layer=None)
        objects = scene.find_by_name('Optimiser')
        optimiserobject = objects[0]
        optimiserobject.update_object_from_optimiser()

    SettingsForm.from_scene(scene, object_types=[OptimiserObject])

    scene.update()
    scene.save()


# ==============================================================================
# Main
# ==============================================================================
if __name__ == '__main__':

    RunCommand(True)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import scriptcontext as sc

import compas_rhino


__commandname__ = "TNO_starting_point"


def RunCommand(is_interactive):

    if 'TNO' not in sc.sticky:
        compas_rhino.display_message('TNO has not been initialised yet.')
        return

    scene = sc.sticky['TNO']['scene']
    proxy = sc.sticky['TNO']['proxy']

    proxy.restart_server()

    objects = scene.find_by_name('Form')
    if not objects:
        compas_rhino.display_message("There is no FormDiagram in the scene.")
        return
    form = objects[0]

    objects = scene.find_by_name('Shape')
    if not objects:
        compas_rhino.display_message("There is no Shape in the scene.")
        return

    # add check to see if loads and

    proxy.package = 'compas_tno.problems'

    formdata = form.diagram.to_data()
    form.diagram.data, output = proxy.initialize_loadpath_proxy(formdata)

    # modify the basic visualisation after analysis
    form.settings['show.reactionvectors'] = True
    form.settings['show.cracks'] = True
    form.settings['show.forcepipes'] = True
    form.settings['show.reactionlabels'] = True

    message = output['status'] + ' - LP: ' + str(round(output['fopt'], 1)) + ' kN.m'

    compas_rhino.display_message(message)

    scene.update()
    scene.save()


# ==============================================================================
# Main
# ==============================================================================
if __name__ == '__main__':

    RunCommand(True)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import scriptcontext as sc

import compas_rhino


__commandname__ = "TNO_run_optimisation"


def RunCommand(is_interactive):

    if 'TNO' not in sc.sticky:
        compas_rhino.display_message('TNO has not been initialised yet.')
        return

    scene = sc.sticky['TNO']['scene']
    proxy = sc.sticky['TNO']['proxy']

    objects = scene.find_by_name('Form')
    if not objects:
        compas_rhino.display_message("There is no FormDiagram in the scene.")
        return
    form = objects[0]

    objects = scene.find_by_name('Shape')
    if not objects:
        compas_rhino.display_message("There is no Shape in the scene.")
        return
    shape = objects[0]

    objects = scene.find_by_name('Optimiser')
    if not objects:
        compas_rhino.display_message("Select First the optimisation settings.")
        return
    optimiser = objects[0]

    # add check to see if loads are applied and if bounds on q are set

    proxy.package = 'compas_tno.problems'

    formdata = form.diagram.to_data()
    shapedata = shape.shape.to_data()
    optimiser.optimiser.message = None
    optimiserdata = optimiser.optimiser.to_data()

    shapedata, formdata, optimiserdata = proxy.run_NLP_proxy(shapedata, formdata, optimiserdata)

    form.diagram.data = formdata
    shape.shape.data = shapedata
    optimiser.optimiser.data = optimiserdata

    message = optimiser.optimiser.message + ' fopt: ' + str(round(optimiser.optimiser.fopt, 2))
    compas_rhino.display_message(message)

    # modify the basic visualisation after analysis
    form.settings['show.reactionvectors'] = True
    form.settings['show.cracks'] = True
    form.settings['show.forcepipes'] = True
    form.settings['show.reactionlabels'] = True

    scene.update()
    scene.save()


# ==============================================================================
# Main
# ==============================================================================
if __name__ == '__main__':

    RunCommand(True)



from compas_tno.shapes import MeshDos
from numpy import array
from scipy import interpolate
import math


__all__ = [
    'interpolate_from_pointcloud',
    'get_shape_ub',
    'get_shape_ub_pattern',
    'get_shape_ub_fill',
    'get_shape_lb',
    'get_shape_lb_pattern',
    'get_shape_middle',
    'delaunay_mesh_from_points',
    'mesh_from_pointcloud',
    'create_mesh_from_topology_and_pointcloud',
    'create_mesh_from_topology_and_basemesh'
]


def interpolate_from_pointcloud(pointcloud, XY, method='linear'):
    pointcloud_array = array(pointcloud)  # TODO: Use instance
    return interpolate.griddata(pointcloud_array[:, :2], pointcloud_array[:, 2], array(XY), method=method)


def get_shape_ub(shape, x, y):
    """Get the height of the extrados in the point.

    Parameters
    ----------
    shape : Shape
        Shape of the masonry
    x : float
        x-coordinate of the point to evaluate.
    y : float
        y-coordinate of the point to evaluate.

    Returns
    -------
    z : float
        The extrados evaluated in the point.
    """
    method = shape.datashape.get('interpolation', 'linear')
    return interpolate_from_pointcloud(shape.extrados.vertices_attributes('xyz'), [x, y], method=method)


def get_shape_ub_pattern(shape, XY):
    """Get the height of the extrados in a list of points.

    Parameters
    ----------
    shape : Shape
        Shape of the masonry
    XY : list or array
        list of the x-coordinate and y-coordinate of the points to evaluate.

    Returns
    -------
    z : float
        The extrados evaluated in the point.
    """
    method = shape.datashape.get('interpolation', 'linear')
    return interpolate_from_pointcloud(shape.extrados.vertices_attributes('xyz'), XY, method=method)


def get_shape_ub_fill(shape, x, y):
    """Get the height of the fill in the point.

    Parameters
    ----------
    shape : Shape
        Shape of the masonry
    x : float
        x-coordinate of the point to evaluate.
    y : float
        y-coordinate of the point to evaluate.

    Returns
    -------
    z : float
        The extrados evaluated in the point.
    """
    method = shape.datashape.get('interpolation', 'linear')
    return interpolate_from_pointcloud(shape.extrados_fill.vertices_attributes('xyz'), [x, y], method=method)


def get_shape_lb(shape, x, y):
    """Get the height of the intrados in the point.

    Parameters
    ----------
    shape : Shape
        Shape of the masonry
    x : float
        x-coordinate of the point to evaluate.
    y : float
        y-coordinate of the point to evaluate.

    Returns
    -------
    z : float
        The intrados evaluated in the point.
    """
    method = shape.datashape.get('interpolation', 'linear')
    return interpolate_from_pointcloud(shape.intrados.vertices_attributes('xyz'), [x, y], method=method)


def get_shape_lb_pattern(shape, XY):
    """Get the height of the intrados in a list of points.

    Parameters
    ----------
    shape : Shape
        Shape of the masonry
    XY : list or array
        list of the x-coordinate and y-coordinate of the points to evaluate.

    Returns
    -------
    z : float
        The extrados evaluated in the point.
    """
    method = shape.datashape.get('interpolation', 'linear')
    return interpolate_from_pointcloud(shape.intrados.vertices_attributes('xyz'), XY, method=method)


def get_shape_middle(shape, x, y):
    """Get the height of the target/middle surface in the point.

    Parameters
    ----------
    shape : Shape
        Shape of the masonry
    x : float
        x-coordinate of the point to evaluate.
    y : float
        y-coordinate of the point to evaluate.

    Returns
    -------
    z : float
        The middle surface evaluated in the point.
    """
    method = shape.datashape.get('interpolation', 'linear')
    return interpolate_from_pointcloud(shape.middle.vertices_attributes('xyz'), [x, y], method=method)


def get_shape_middle_pattern(shape, XY):
    """Get the height of the target/middle surface in a list of points.

    Parameters
    ----------
    shape : Shape
        Shape of the masonry
    XY : list or array
        list of the x-coordinate and y-coordinate of the points to evaluate.

    Returns
    -------
    z : float
        The extrados evaluated in the point.
    """
    method = shape.datashape.get('interpolation', 'linear')
    return interpolate_from_pointcloud(shape.middle.vertices_attributes('xyz'), XY, method=method)


def delaunay_mesh_from_points(points):
    """Construct a Delaunay triangulation of set of vertices.

    Parameters
    ----------
    points : list
        XY(Z) coordinates of the points to triangulate.

    Returns
    -------
    mesh : MeshDos
        Delaunay mesh created.

    """

    from triangle import triangulate  # check to do it with scipy
    data = {'vertices': [point[0:2] for point in points]}
    result = triangulate(data, opts='c')
    vertices = []
    vertices_flat = []
    i = 0
    for x, y in result['vertices']:
        vertices.append([x, y, points[i][2]])
        vertices_flat.append([x, y, 0.0])
        i += 1
    faces = result['triangles']
    faces_meaningful = []

    mesh_flat = MeshDos.from_vertices_and_faces(vertices_flat, faces)

    i = 0
    for fkey in mesh_flat.faces():  # Did this to avoid faces with area = 0, see if it is necessary
        if mesh_flat.face_area(fkey) > 0.001:
            faces_meaningful.append(faces[i])
        i = i+1

    mesh = MeshDos.from_vertices_and_faces(vertices, faces_meaningful)

    return mesh


def mesh_from_pointcloud(points):
    """Construct a Delaunay triangulation of set of vertices.

    Parameters
    ----------
    points : list
        XY(Z) coordinates of the points to triangulate.

    Returns
    -------
    mesh : MeshDos
        Delaunay mesh created.

    """

    mesh = delaunay_mesh_from_points(points)
    XY = mesh.vertices_attributes('xy')
    z = interpolate_from_pointcloud(points, XY)

    for i, key in enumerate(mesh.vertices()):
        mesh.vertex_attribute(key, 'z', z[i])

    return mesh


def create_mesh_from_topology_and_pointcloud(meshtopology, pointcloud, isnan_height=0.0):
    """
    Create a mesh based on a given topology and the heights based in a pointcloud.

    Parameters
    ----------
    meshtopology : FormDiagram
        Topology that is intended to keep
    mesh_base : mesh
        Mesh usually denser to use as base to interpolate the heights
    isnan_height : bool, optional
        Value if the height is nan, by default is 0.0

    Returns
    -------
    obj
        Mesh.

    """
    vertices, faces = meshtopology.to_vertices_and_faces()
    mesh = MeshDos.from_vertices_and_faces(vertices, faces)
    XY = mesh.vertices_attributes('xy')
    z = interpolate_from_pointcloud(pointcloud, XY)

    for i, key in enumerate(mesh.vertices()):
        if math.isnan(z[i]):
            print('Height (nan) for [x,y]:', XY[i])
            z[i] = isnan_height
        mesh.vertex_attribute(key, 'z', float(z[i]))

    return mesh


def create_mesh_from_topology_and_basemesh(meshtopology, mesh_base):
    """
    Create a mesh based on a given topology and the heights based in a base mesh (usually denser).

    Parameters
    ----------
    meshtopology : compas_tno.diagrams.FormDiagram
        Topology that is intended to keep
    mesh_base : mesh
        Mesh usually denser to use as base to interpolate the heights

    Returns
    -------
    obj
        Mesh.

    """
    vertices, faces = meshtopology.to_vertices_and_faces()
    mesh = MeshDos.from_vertices_and_faces(vertices, faces)
    XY = mesh.vertices_attributes('xy')
    XYZ_base = mesh_base.vertices_attributes('xyz')
    z = interpolate_from_pointcloud(XYZ_base, XY)

    for i, key in enumerate(mesh.vertices()):
        mesh.vertex_attribute(key, 'z', float(z[i]))

    return mesh



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from functools import partial
from compas.utilities import color_to_colordict
from compas_rhino.artists import MeshArtist


colordict = partial(color_to_colordict, colorformat='rgb', normalize=False)


__all__ = ['DiagramArtist']


class DiagramArtist(MeshArtist):
    """Base artist for diagrams in TNO.

    Attributes
    ----------
    diagram : :class:`compas_tno.diagrams.Diagram`
        The diagram associated with the artist.

    """

    @property
    def diagram(self):
        """The diagram assigned to the artist."""
        return self.mesh

    @diagram.setter
    def diagram(self, diagram):
        self.mesh = diagram




from numpy import hstack
from numpy import array
from numpy import asarray
from numpy import any
from numpy.linalg import matrix_rank
from numpy.linalg import svd
from numpy.random import rand
from numpy import diag
from scipy.linalg import qr
from math import sqrt


def find_independents_forward(E, nind=None, tol=None):
    """ Find independent edges of the matrix E with the forward method.
    The matrix E is reconstructed column by column and the rank is computed at each step.
    Everytime that a column is added and the rank does not increase, the edge corresp. that column is selected as independent.

    Parameters
    ----------
    E : array
        Equilibrium matrix.
    nind : ind
        Number of independent to stop the process, if known in advance, by default None.
    tol : float
        Tolerance for small Singular Values. Default is None.

    Returns
    -------
    ind : list
        Independent columns.

    """

    ind = []
    _, m = E.shape

    # Take as independent all edges connecting supported points.

    for i in range(m):
        if any(E[:, i]):
            pass
        else:
            ind.append(i)

    internal = list(set(range(m)) - set(ind))
    Etemp = E[:, [internal[0]]]

    # Proceed for the internal edges only.

    for i in internal[1:]:
        Etest = hstack([Etemp, E[:, [i]]])
        _, ncol = Etest.shape
        if matrix_rank(Etest, tol=tol) < ncol:  # testing tol
            ind.append(i)
        else:
            Etemp = Etest
        if nind:
            if len(ind) == nind:
                break

    return ind


def find_independents_backward(E, nind=None, tol=None):
    """ Find independent edges of the matrix E with the backward method.
    The last columns of the matrix are removed sequentially and the rank is checked.
    Everytime thae rank does not decrease, the edge is selected as independent.

    Parameters
    ----------
    E : array
        Equilibrium matrix.
    nind : ind
        Number of independent to stop the process, if known in advance, by default None.
    tol : float
        Tolerance for small singular values. Default is None.

    Returns
    -------
    ind : list
        Independent columns.

    """

    ind = []
    _, m = E.shape

    # Take as independent all edges connecting supported points.

    for i in range(m):
        if any(E[:, i]):
            pass
        else:
            ind.append(i)

    internal = list(set(range(m)) - set(ind))
    rank = matrix_rank(E, tol=tol)

    for i in range(m):
        i_inv = m - i - 1
        indices_red = [j for j in internal if not j == i_inv]
        Ered = E[:, indices_red]
        ri = matrix_rank(Ered, tol=tol)
        if ri == rank:  # belongs to the nullspace
            ind.append(i_inv)
            internal.remove(i_inv)
            rank = ri
        if nind:
            if len(ind) == nind:
                break

    return ind


def find_independents_QR(E, tol=None):
    """ Find independent edges of the matrix E based on a permuted QR factorization.
    The matrix E [m x n] (n > m) is factored in Q, R, P matrices, where Q [m x m] is a square matrix, R [m x n] an upper triangular
    matrix and P is the column permutation applied.
    This method is faster, but it does not preserve the numbering structure of the form diagram.

    Parameters
    ----------
    E : array
        Equilibrium matrix.
    tol : float
        Tolerance for small singular values. Default is None.

    Returns
    -------
    ind : list
        Independent columns.

    """

    if not tol:
        tol = 1e-12

    _, R, P = qr(E, pivoting=True)

    dr = abs(diag(R))
    rank_ = sum(dr > tol*dr[0])
    ind = sorted(list(P[rank_:]))

    return ind


def find_independents(E, method='SVD', tol=None):
    """ Overall method to find independent edges dependent on method

    Parameters
    ----------
    E : array
        Equilibrium matrix.
    method : str
        Method to find independents. Default is 'SVD'. More to come.
    tol : float
        Tolerance for small singular values. Default is None.

    Returns
    -------
    ind : list
        Independent columns.

    """

    if method == 'SVD':
        ind = find_independents_forward(E, tol)
    else:
        raise ValueError('Plese select a valid method to find the independent edges')

    return ind


def independents_exclude(E, outs):
    """ Find inndependent edges with certain to exclude.

    Parameters
    ----------
    E : array
        Equilibrium matrix.
    outs : list
        List of edges (columns in the matrix) to exclude.

    Returns
    -------
    ind : list
        Independent columns.

    """

    _, m = E.shape
    possible = list(set(range(m)) - set(outs))
    Eouts = E[:, outs]
    if matrix_rank(Eouts) == Eouts.shape[1]:
        Etemp = Eouts
        ind = []
    else:
        print('Warning, could not exclude all')
        return find_independents_forward(E)

    for i in possible:
        Etest = hstack([Etemp, E[:, [i]]])
        _, ncol = Etest.shape
        if matrix_rank(Etest) < ncol:
            ind.append(i)
        else:
            Etemp = Etest

    return ind


def independents_include(E, ins):
    """ Find inndependent edges with certain to include.

    Parameters
    ----------
    E : array
        Equilibrium matrix.
    ins : list
        List of edges (columns in the matrix) to include.

    Returns
    -------
    ind : list
        Independent columns.


    """

    n, m = E.shape
    if len(ins) > (m-n):
        print('Too many included edges - limit to number of independents: {0}'.format((m-n)))
        ins = ins[:(m-n)]
    not_in = list(set(range(m)) - set(ins))
    Ein = E[:, ins]
    while matrix_rank(Ein) < Ein.shape[1]:
        print('Warning, edges are dependent among each other')
        ins = ins[matrix_rank(Ein)]
    if len(ins) == (m-n):
        Enot_in = E[:, not_in]
        if matrix_rank(Enot_in) == Enot_in.shape[1]:
            return ins
        else:
            print('Warning, edges do not form an independent set')
            return find_independents_forward(E)
    ind = ins
    Etemp = E[:, [not_in[0]]]
    Etemp.shape

    for i in not_in[1:]:
        Etest = hstack([Etemp, E[:, [i]]])
        _, ncol = Etest.shape
        if matrix_rank(Etest) < ncol:
            ind.append(i)
        else:
            Etemp = Etest

    return ind


def inds_incl_excl(E, ins, outs):
    """ Find inndependent edges with certain to exclude and include.

    Parameters
    ----------
    E : array
        Equilibrium matrix.
    ins : list
        List of edges (columns in the matrix) to include.
    outs : list
        List of edges (columns in the matrix) to exclude.

    Returns
    -------
    ind : list
        Independent columns.

    """

    n, m = E.shape
    if len(ins) > (m-n):
        print('Too many included edges - limit to number of independents: {0}'.format((m-n)))
        ins = ins[:(m-n)]
    not_in = list(set(range(m)) - set(ins))
    possible = list(set(not_in)-set(outs))
    Ein = E[:, ins]
    Eouts = E[:, outs]
    while matrix_rank(Ein) < Ein.shape[1]:
        print('Warning, included edges are dependent among each other')
        ins = ins[matrix_rank(Ein)]
    if len(ins) == (m-n):
        Enot_in = E[:, not_in]
        if matrix_rank(Enot_in) == Enot_in.shape[1]:
            return ins
        else:
            print('Warning, edges do not form an independent set')
            return find_independents_forward(E)
    if matrix_rank(Eouts) == Eouts.shape[1]:
        Etemp = Eouts
        ind = ins
    else:
        print('Warning, could not exclude all')
        ind = ins
        Etemp = E[:, [not_in[0]]]
        possible = not_in[1:]

    for i in possible:
        Etest = hstack([Etemp, E[:, [i]]])
        _, ncol = Etest.shape
        if matrix_rank(Etest) < ncol:
            ind.append(i)
        else:
            Etemp = Etest

    return ind


def check_independents(M, tol=0.001):
    """ Run checks to verify the independennts.

    Parameters
    ----------
    M : :class:`~compas_tno.problems.Problem`
        Arguments from the optimisation.
    tol : float, optional
        Allowed error.
        The default values is ``0.001``.

    Returns
    -------
    checked : bool
        True if independents are checked.

    """

    checked = True
    if tol > 0:
        for i in range(10**2):
            qid = array(rand(M.k) * 10).reshape(-1, 1)
            q_ = M.B @ qid + M.d
            res_x = M.Citx.dot(M.U * q_.ravel()) - M.ph[:len(M.free_x)].ravel()
            res_y = M.City.dot(M.V * q_.ravel()) - M.ph[:len(M.free_y)].ravel()
            R = max(sqrt(max(res_x**2)), sqrt(max(res_y**2)))
            if R > tol:
                checked = False
                print('deviation exceeded limit:', R)
                break

    # checked = True
    # q_ = deepcopy(M.q)
    # if tol > 0:
    #     for i in range(10**3):
    #         q_[M.ind, 0] = rand(M.k) * 10
    #         q_[M.dep] = M.Edinv.dot(- M.ph + M.Ei.dot(q_[M.ind]))
    #         Rx = array(M.Citx.dot(M.U * q_.ravel()) - M.ph[:len(M.free_x)].ravel())
    #         Ry = array(M.City.dot(M.V * q_.ravel()) - M.ph[len(M.free_y):].ravel())
    #         R = max(sqrt(max(Rx**2)), sqrt(max(Ry**2)))
    #         # print(R)
    #         if R > tol:
    #             checked = False
    #             break

    return checked


def check_horizontal_loads(E, p):
    """ Run checks to verify of a certain horizontal distribution can be taken by the fixed form diagram.

    Parameters
    ----------
    E : array
        Equilibrium matrix.
    p : array
        Vector of horizontal loads.

    Returns
    -------
    checked : bool
        True if independents are checked.

    """

    r = matrix_rank(E)
    r_ = matrix_rank(hstack([E, p]))

    if r == r_:
        checked = True
    else:
        checked = False

    return checked


def count_inds(E, tol=None):
    """ Count the number of independent edges for a given singular value tolerance.

    Parameters
    ----------
    E : array
        Equilibrium matrix.
    tol : float
        Tolerance for small Singular Values. Default is None.

    Returns
    -------
    k : int
        number of independent edges
    rd : int
        Rank Deficiency of the Equilibrium Matrix.

    """

    nl, ncol = E.shape
    k = max(nl, ncol) - matrix_rank(E, tol=tol)
    rd = k - abs(nl - ncol)

    return k, rd


def matrix_svd(E):
    """ Return the singular values of the matrix E.

    Parameters
    ----------
    E : array
        Matrix to compute the singular values.

    Returns
    -------
    s : array
        vector with the singular values/
    """

    _, s, _ = svd(asarray(E))

    return s



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import compas_rhino
from compas_rhino.artists import BaseArtist
from compas.utilities import pairwise
from compas.geometry import centroid_polygon


__all__ = ['ShapeArtist']


class ShapeArtist(BaseArtist):
    """Artist for shape in TNO.

    Parameters
    ----------
    shape: compas_tno.shapes.Shape
        The shape to draw.

    """

    def __init__(self, shape, scene=None, name=None, layer=None, visible=True, settings=None):
        super(ShapeArtist, self).__init__()
        self._shape = shape
        self.intrados_layer = "Shape::Intrados"
        self.extrados_layer = "Shape::Extrados"
        self.middle_layer = "Shape::Middle"

    @property
    def shape(self):
        """The shape associated with the artist."""
        return self._shape

    @shape.setter
    def shape(self, shape):
        self._shape = shape

    def draw_intrados(self, color=(0, 0, 0), displacement=None, layer=None):

        layer = layer or self.intrados_layer
        intrados = self.shape.intrados

        return self.draw_shape_mesh(intrados, color=color, layer=layer)

    def draw_extrados(self, color=(0, 0, 0), displacement=None, layer=None):

        layer = layer or self.extrados_layer
        extrados = self.shape.extrados

        return self.draw_shape_mesh(extrados, color=color, layer=layer)

    def draw_middle(self, color=(0, 0, 0), displacement=None, layer=None):

        layer = layer or self.middle_layer
        middle = self.shape.middle

        return self.draw_shape_mesh(middle, color=color, layer=layer)

    def draw_shape_mesh(self, mesh, color=(0, 0, 0), disjoint=False, layer=None):
        """Draw the meshes of the shape object as a consolidated RhinoMesh.

        Parameters
        ----------
        color : tuple, optional
            The color of the mesh.
            Default is black, ``(0, 0, 0)``.
        disjoint : bool, optional
            Draw the faces of the mesh with disjoint vertices.
            Default is ``False``.

        Returns
        -------
        list
            The GUIDs of the created Rhino objects.

        Notes
        -----
        The mesh should be a valid Rhino Mesh object, which means it should have only triangular or quadrilateral faces.
        Faces with more than 4 vertices will be triangulated on-the-fly.

        """
        vertex_index = mesh.key_index()
        vertices = [mesh.vertex_coordinates(vertex) for vertex in mesh.vertices()]
        faces = [[vertex_index[vertex] for vertex in mesh.face_vertices(face)] for face in mesh.faces()]
        new_faces = []
        for face in faces:
            f = len(face)
            if f == 3:
                new_faces.append(face + face[-1:])
            elif f == 4:
                new_faces.append(face)
            elif f > 4:
                centroid = len(vertices)
                vertices.append(centroid_polygon([vertices[index] for index in face]))
                for a, b in pairwise(face + face[0:1]):
                    new_faces.append([centroid, a, b, b])
            else:
                continue
        name = "{}".format(mesh.name)
        guid = compas_rhino.draw_mesh(vertices, new_faces, layer=layer, name=name, color=color, disjoint=disjoint)
        return [guid]



"""
********************************************************************************
compas_tno.diagrams
********************************************************************************

.. currentmodule:: compas_tno.diagrams


Classes
=======

.. autosummary::
    :toctree: generated/

    FormDiagram
    ForceDiagram
    FormGraph


Rectangular diagrams
====================

.. autosummary::
    :toctree: generated/

    create_cross_form
    create_cross_diagonal
    create_cross_with_diagonal
    create_fan_form
    create_ortho_form
    create_parametric_form


Circular diagrams
=================

.. autosummary::
    :toctree: generated/

    create_circular_radial_form
    create_circular_radial_spaced_form
    create_circular_spiral_form


Linear diagrams
===============

.. autosummary::
    :toctree: generated/

    create_arch_form_diagram
    create_linear_form_diagram

"""
from __future__ import absolute_import

from .force import ForceDiagram
from .form import FormDiagram
from .graph import FormGraph

from .diagram_arch import (
    create_arch_form_diagram,
    create_linear_form_diagram
)
from .diagram_circular import (
    create_circular_radial_form,
    create_circular_radial_spaced_form,
    create_circular_spiral_form,
)
from .diagram_rectangular import (
    create_cross_form,
    create_cross_diagonal,
    create_cross_with_diagonal,
    create_fan_form,
    create_ortho_form,
    create_parametric_form
)

__all__ = [
    'ForceDiagram',
    'FormDiagram',
    'FormGraph',

    'create_arch_form_diagram',
    'create_linear_form_diagram',

    'create_circular_radial_form',
    'create_circular_radial_spaced_form',
    'create_circular_spiral_form',

    'create_cross_form',
    'create_cross_diagonal',
    'create_cross_with_diagonal',
    'create_fan_form',
    'create_ortho_form',
    'create_parametric_form'

]



from numpy import zeros

from compas.geometry import norm_vector

import math


def general_ub_lb_update_with_t_middle_constant(thk, s, middle, t):
    """Update upper and lower bounds for a general surface for a given middle surface.

    Parameters
    ----------
    thk : float
        Thickness of the structure
    s : float
        Middle surface heights
    middle : :class:`~compas_tno.shapes.MeshDos`
        Middle surface mesh
    t : float
        Value to assumed to vertices without intrados projection

    Returns
    -------
    ub_update : array
        Upper bound limits updated
    lb_update : array
        Lower bound limits updated
    """

    ub_update = zeros((len(s), 1))
    lb_update = zeros((len(s), 1))

    i = 0
    for key in middle.vertices():
        normal = middle.vertex_attribute(key, 'n')
        dev = 1/math.sqrt(1/(1 + (normal[0]**2 + normal[1]**2)/normal[2]**2))  # 1/cos(a)
        ub_update[i] = s[i] + thk/2 * dev * norm_vector(normal)  # Experimenting with this normal norm!
        if middle.vertex_attribute(key, 'is_outside'):
            lb_update[i] = t
        else:
            lb_update[i] = s[i] - thk/2 * dev * norm_vector(normal)
        i += 1

    return ub_update, lb_update


def general_db_with_t_middle_constant(s, middle):
    """Update sensitivity in the upper and lower bounds for a general surface for a given middle surface.

    Parameters
    ----------
    s : float
        Middle surface heights
    middle : MeshDos
        Middle surface mesh

    Returns
    -------
    dub : array
        Sensitivities of the upper bound
    dlb : array
        Sensitivities of the lower bound
    """

    dub = zeros((len(s), 1))
    dlb = zeros((len(s), 1))

    i = 0
    for key in middle.vertices():
        normal = middle.vertex_attribute(key, 'n')
        dev = 1/math.sqrt(1/(1 + (normal[0]**2 + normal[1]**2)/normal[2]**2))  # 1/cos(a)
        dub[i] = 1/2 * dev * norm_vector(normal)  # Experimenting with this normal norm!
        if middle.vertex_attribute(key, 'is_outside'):
            dlb[i] = 0.0
        else:
            dlb[i] = - 1/2 * dev * norm_vector(normal)
        i += 1

    return dub, dlb


def general_ub_lb_update_with_t_middle_variable(thk_alfa, s, middle, t):
    """Update upper and lower bounds for a general surface for a given middle surface. This general  update takes into consideration nub and nlb. scaled, so t is a (%)

    Parameters
    ----------
    thk_alfa : float
        Thickness of the structure as a percentage of nub, nlb
    s : float
        Middle surface heights
    middle : :class:`~compas_tno.shapes.MeshDos`
        Middle surface mesh
    t : float
        Value to assumed to vertices without intrados projection

    Returns
    -------
    ub_update : array
        Upper bound limits updated
    lb_update : array
        Lower bound limits updated
    """

    ub_update = zeros((len(s), 1))
    lb_update = zeros((len(s), 1))

    i = 0
    for key in middle.vertices():
        # normal = middle.vertex_attribute(key, 'n')
        nub = middle.vertex_attribute(key, 'nub')
        nlb = middle.vertex_attribute(key, 'nlb')
        dev_ub = 1/math.sqrt(1/(1 + (nub[0]**2 + nub[1]**2)/nub[2]**2))  # 1/cos(a)
        if middle.vertex_attribute(key, 'is_outside'):
            dev_lb = 0.0
        else:
            dev_lb = 1/math.sqrt(1/(1 + (nlb[0]**2 + nlb[1]**2)/nlb[2]**2))  # 1/cos(a)
        ub_update[i] = + thk_alfa * dev_ub * norm_vector(nub)  # Experimenting with this normal norm!
        lb_update[i] = - thk_alfa * dev_lb * norm_vector(nlb)  # check if neeed to divide by 2
        # if intrados.vertex_attribute(key, 'is_outside'):
        #     lb_update[i] = t
        i += 1

    return ub_update + s, lb_update + s


def general_db_with_t_middle_variable(s, middle):
    """Update sensitivity in the upper and lower bounds for a general surface for a given middle surface considering the variable as a percentage.

    Parameters
    ----------
    s : float
        Middle surface heights
    middle : :class:`~compas_tno.shapes.MeshDos`
        Middle surface mesh

    Returns
    -------
    dub : array
        Sensitivities of the upper bound
    dlb : array
        Sensitivities of the lower bound
    """

    dub = zeros((len(s), 1))
    dlb = zeros((len(s), 1))

    i = 0
    for key in middle.vertices():
        # normal = middle.vertex_attribute(key, 'n')
        nub = middle.vertex_attribute(key, 'nub')
        nlb = middle.vertex_attribute(key, 'nlb')
        dev_ub = 1/math.sqrt(1/(1 + (nub[0]**2 + nub[1]**2)/nub[2]**2))  # 1/cos(a)
        if middle.vertex_attribute(key, 'is_outside'):
            dev_lb = 0.0
        else:
            dev_lb = 1/math.sqrt(1/(1 + (nlb[0]**2 + nlb[1]**2)/nlb[2]**2))  # 1/cos(a)
        dub[i] = + 1 * dev_ub * norm_vector(nub)
        dlb[i] = - 1 * dev_lb * norm_vector(nlb)
        i += 1

    return dub, dlb


def general_ub_lb_update_with_t_intrados(thk, lb, intrados, t):
    """Update upper and lower bounds for a general surface for a given lower bound (intrados) surface.

    Parameters
    ----------
    thk_alfa : float
        Thickness of the structure as a percentage of nub, nlb
    s : float
        Middle surface heights
    middle : :class:`~compas_tno.shapes.MeshDos`
        Middle surface mesh
    t : float
        Value to assumed to vertices without intrados projection

    Returns
    -------
    ub_update : array
        Upper bound limits updated
    lb_update : array
        Lower bound limits updated
    """

    ub_update = zeros((len(lb), 1))
    lb_update = lb

    i = 0
    for key in intrados.vertices():
        normal = intrados.vertex_attribute(key, 'n')
        dev = 1/math.sqrt(1/(1 + (normal[0]**2 + normal[1]**2)/normal[2]**2))  # 1/cos(a)
        ub_update[i] = lb[i] + 1.0 * thk * dev * norm_vector(normal)  # Experimenting with this normal norm!
        i += 1

    return ub_update, lb_update


def general_db_with_t_intrados(lb, intrados):
    """Update sensitivity in the upper and lower bounds for a general surface for a given lower bound (intrados) surface.

    Parameters
    ----------
    s : float
        Middle surface heights
    middle : :class:`~compas_tno.shapes.MeshDos`
        Middle surface mesh

    Returns
    -------
    dub : array
        Sensitivities of the upper bound
    dlb : array
        Sensitivities of the lower bound
    """

    dub = zeros((len(lb), 1))
    dlb = zeros((len(lb), 1))

    i = 0
    for key in intrados.vertices():
        normal = intrados.vertex_attribute(key, 'n')
        dev = 1/math.sqrt(1/(1 + (normal[0]**2 + normal[1]**2)/normal[2]**2))  # 1/cos(a)
        dub[i] = 1 * dev * norm_vector(normal)  # Experimenting with this normal norm!
        i += 1

    return dub, dlb


def general_ub_lb_update_with_s(ub, lb, s):
    """Update upper and lower bounds for a general surface based on upper and lower bounds that are squized by the ``s`` variable.

    Parameters
    ----------
    ub : array
        Upper bound original limits
    lb : array
        Lower bound original limits
    s : float
        Parameter to reduce the thickness (as a percentual of the original section)

    Returns
    -------
    ub_update : array
        Upper bound limits updated
    lb_update : array
        Lower bound limits updated
    """

    ub_update = zeros((len(ub), 1))
    lb_update = zeros((len(ub), 1))

    for i in range(len(ub)):
        ub_update[i] = ub[i] - (ub[i] - lb[i]) * s
        lb_update[i] = lb[i] + (ub[i] - lb[i]) * s

    return ub_update, lb_update


def general_dub_dlb_with_s(ub, lb):
    """Update sensitivity in the upper and lower bounds for a general surface for a given lower bound (intrados) surface.

    Parameters
    ----------
    ub : array
        Current bound limits
    lb : array
        Current bound limits

    Returns
    -------
    dub : array
        Sensitivities of the upper bound
    dlb : array
        Sensitivities of the lower bound
    """

    dub = zeros((len(ub), 1))
    dlb = zeros((len(ub), 1))

    for i in range(len(ub)):
        dub[i] = - (ub[i] - lb[i])
        dlb[i] = + (ub[i] - lb[i])

    return dub, dlb


def general_ub_lb_update_with_n(ub, lb, n, intrados, extrados, t):
    """Update upper and lower bounds for a general surface for two given upper and lower bound surfaces.

    Parameters
    ----------
    ub : array
        Current bound limits
    lb : array
        Current bound limits
    n : float
        Magnitude of the offset from the original bounds
    intrados : MeshDos
        Intrados surface mesh
    extrados : MeshDos
        Extrados surface mesh
    t : float
        Value to assumed to vertices without intrados projection

    Returns
    -------
    ub_update : array
        Upper bound limits updated
    lb_update : array
        Lower bound limits updated
    """

    ub_update = zeros((len(ub), 1))
    lb_update = zeros((len(ub), 1))

    # Intrados and Extrados must me diagrams with exactly same topology as the form diagram: i.e. projections of the form diagram with z = ub,lb.

    i = 0
    for key in intrados.vertices():
        normal_intra = intrados.vertex_attribute(key, 'n')
        normal_extra = extrados.vertex_attribute(key, 'n')

        deviation_extra = 1/math.sqrt(1/(1 + (normal_extra[0]**2 + normal_extra[1]**2)/normal_extra[2]**2))  # 1/cos(a)
        ub_update[i] = ub[i] - n * deviation_extra * norm_vector(normal_extra)  # Experimenting with this normal norm!

        if intrados.vertex_attribute(key, 'is_outside'):
            lb_update[i] = t
        else:
            deviation_intra = 1/math.sqrt(1/(1 + (normal_intra[0]**2 + normal_intra[1]**2)/normal_intra[2]**2))  # 1/cos(a)
            lb_update[i] = lb[i] + n * deviation_intra * norm_vector(normal_intra)  # Experimenting with this normal norm!
        i += 1

    return ub_update, lb_update


def general_dub_dlb_with_n(ub, lb, n, intrados, extrados, t):
    """Sensitivity upper and lower bounds for a general surface for two given upper and lower bound surfaces.

    Parameters
    ----------
    ub : array
        Current bound limits
    lb : array
        Current bound limits
    n : float
        Magnitude of the offset from the original bounds
    intrados : :class:`~compas_tno.shapes.MeshDos`
        Intrados surface mesh
    extrados : :class:`~compas_tno.shapes.MeshDos`
        Extrados surface mesh
    t : float
        Value to assumed to vertices without intrados projection

    Returns
    -------
    dub : array
        Sensitivity of upper bound limits
    dlb : array
        Sensitivity of lower bound limits
    """

    dub = zeros((len(ub), 1))
    dlb = zeros((len(ub), 1))

    i = 0
    for key in intrados.vertices():
        normal_intra = intrados.vertex_attribute(key, 'n')
        normal_extra = extrados.vertex_attribute(key, 'n')
        x, y, z_lb = intrados.vertex_coordinates(key)
        x, y, z_ub = extrados.vertex_coordinates(key)
        if intrados.vertex_attribute(key, 'is_outside'):
            deviation_intra = 0.0
        else:
            deviation_intra = 1/math.sqrt(1/(1 + (normal_intra[0]**2 + normal_intra[1]**2)/normal_intra[2]**2))  # 1/cos(a)
        deviation_extra = 1/math.sqrt(1/(1 + (normal_extra[0]**2 + normal_extra[1]**2)/normal_extra[2]**2))  # 1/cos(a)
        dlb[i] = + deviation_intra * norm_vector(normal_intra)  # Experimenting with this normal norm!
        dub[i] = - deviation_extra * norm_vector(normal_extra)  # Experimenting with this normal norm!
        i += 1

    return dub, dlb


def _invert_vector(vector):
    return [-1*vector[0], -1*vector[1], -1*vector[2]]


def general_b_update_with_n(b, n, fixed):
    return


def general_db_with_n(b, n, fixed):

    return



from compas_tno.solvers import mma_numpy
from numpy import hstack
from numpy import array
from numpy import zeros

from .post_process import post_process_general

import time

from compas_tno.problems import sensitivities_wrapper
from compas_tno.problems import constr_wrapper

try:
    from torch import tensor

    from compas_tno.autodiff.equilibrium_pytorch import f_constraints_pytorch_MMA
    from compas_tno.autodiff.equilibrium_pytorch import f_objective_pytorch
    from compas_tno.autodiff.equilibrium_pytorch import compute_autograd
    from compas_tno.autodiff.equilibrium_pytorch import compute_autograd_jacobian
except BaseException:
    pass


def run_optimisation_MMA(analysis):
    """ Run convex optimisation problem with MMA.

    Parameters
    ----------
    obj : analysis
        Analysis object with information about optimiser, form and shape.

    Returns
    -------
    obj : analysis
        Analysis object optimised.

    """

    optimiser = analysis.optimiser
    solver = optimiser.settings['solver']
    constraints = optimiser.settings['constraints']
    objective = optimiser.settings['objective']
    gradient = optimiser.settings.get('gradient', False)
    jacobian = optimiser.settings.get('jacobian', False)
    fconstr = optimiser.fconstr
    args = optimiser.args

    bounds = optimiser.bounds
    x0 = optimiser.x0
    g0 = optimiser.g0

    if solver != 'MMA':
        print('Error, Only MMA solver is available for this library!')

    fobj = optimiser.fobj
    fgrad = optimiser.fgrad
    fjac = optimiser.fjac
    fconstr = optimiser.fconstr

    start_time = time.time()

    if gradient and jacobian:

        import nlopt

        # opt = nlopt.opt(nlopt.LD_MMA, len(x0))
        opt = nlopt.opt(nlopt.LD_SLSQP, len(x0))
        fobj_nlopt = create_nlopt_fobj(fobj, fgrad, *args)
        fconstr_nlopt = create_nlopt_fconstr(fconstr, fjac, *args)

        opt.set_min_objective(fobj_nlopt)
        lower = array([lw[0] for lw in bounds]).flatten()
        upper = array([up[1] for up in bounds]).flatten()
        opt.set_lower_bounds(lower)
        opt.set_upper_bounds(upper)
        opt.add_inequality_mconstraint(fconstr_nlopt, zeros((len(g0),)))
        opt.set_maxtime(1000)
        opt.set_xtol_rel(1e-8)

        xopt = opt.optimize(x0.flatten())
        fopt = opt.last_optimum_value()
        result = opt.last_optimize_result()
        if result > 0:
            exitflag = 0

        # args_MMA = list(args)
        # args_MMA.append([fobj, fgrad])
        # f_g_eval = analytical_f_g
        # f_g_df_dg_eval = analytical_f_g_df_dg
    else:

        (q, ind, dep, E, Edinv, Ei, C, Ct, Ci, Cit, Cf, U, V, p, px, py, pz, z, free, fixed, lh, sym, k, lb, ub, lb_ind, ub_ind, s, Wfree, x, y, b, joints,
         cracks_lb, cracks_ub, free_x, free_y, rol_x, rol_y, Citx, City, Cftx, Cfty, qmin, constraints, max_rol_rx, max_rol_ry, Asym) = args[:48]

        args_MMA = list(args)
        args_MMA.append([fobj, fconstr])
        f_g_eval = pytorch_f_g
        f_g_df_dg_eval = pytorch_f_g_df_dg
        EdinvEi = Edinv*Ei
        Edinv_p = Edinv.dot(p)

        EdinvEi_th = tensor(EdinvEi)
        Edinv_p_th = tensor(Edinv_p)
        C_th = tensor(C.toarray())
        Ci_th = tensor(Ci.toarray())
        Cit_th = Ci_th.t()
        Cf_th = tensor(Cf.toarray())
        pzfree = tensor(pz[free])
        xyz = tensor(hstack([x, y, z]))
        xy = tensor(hstack([x, y]))
        pfixed = tensor(hstack([px, py, pz])[fixed])
        U_th = tensor(U.toarray())
        V_th = tensor(V.toarray())

        args_obj = (Edinv_p_th, EdinvEi_th, ind, dep, C_th, Ci_th, Cit_th, Cf_th, pzfree, xyz, xy, pfixed, k, objective)
        args_constr = (Edinv_p_th, EdinvEi_th, ind, dep, C_th, Ci_th, Cit_th, Cf_th, pzfree, xyz, xy, pfixed, k, free, fixed,
                       ub, lb, ub_ind, lb_ind, b, constraints, max_rol_rx, max_rol_ry, rol_x, rol_y, px, py, Asym, U_th, V_th)

        args_MMA = [args_obj, args_constr]
        fopt, xopt, exitflag = mma_numpy(f_g_eval, f_g_df_dg_eval, x0, bounds, args_MMA, 10e-4, 100)

    elapsed_time = time.time() - start_time
    optimiser.exitflag = exitflag
    optimiser.time = elapsed_time
    optimiser.fopt = fopt
    optimiser.xopt = xopt
    optimiser.niter = None  # Did not find a way to display number of iteratinos
    if exitflag == 0:
        optimiser.message = 'Optimisation succesful terminated'  # Did not find a way to display message
    else:
        optimiser.message = 'Optimal solution not found'

    post_process_general(analysis)

    return analysis


# ---------------------------------------------------------------
# Calulation with pytorch
# ---------------------------------------------------------------


def pytorch_f_g(variables, *args):
    variables = tensor(variables, requires_grad=False)
    args_obj = args[0]
    args_constr = args[1]
    f0val = array(f_objective_pytorch(variables, *args_obj))
    fval = - 1 * array(f_constraints_pytorch_MMA(variables, *args_constr))
    return f0val, fval


def pytorch_f_g_df_dg(variables, *args):
    variables = tensor(variables, requires_grad=True)
    args_obj = args[0]
    args_constr = args[1]
    fval_tensor = -1 * f_constraints_pytorch_MMA(variables, *args_constr)
    dfdx = array(compute_autograd_jacobian(variables, fval_tensor))
    f0val_tensor = f_objective_pytorch(variables, *args_obj)
    df0dx = array(compute_autograd(variables, f0val_tensor))
    f0val = f0val_tensor.detach().numpy()
    fval = fval_tensor.detach().numpy()
    return f0val, df0dx, fval, dfdx


# ---------------------------------------------------------------
# Calculation analytical
# ---------------------------------------------------------------


def analytical_f_g(xopt, *args):
    [fobj, _] = args[-1]
    args = args[:-1]
    f0val = fobj(xopt, *args)
    fval = - 1 * constr_wrapper(xopt, *args)  # change args here
    return f0val, fval.reshape(-1, 1)


def analytical_f_g_df_dg(xopt, *args):
    [fobj, fgrad] = args[-1]
    args = args[:-1]
    f0val = fobj(xopt, *args)
    fval = - 1 * constr_wrapper(xopt, *args)  # change args here
    df0dx = fgrad(xopt, *args)
    dfdx = -1 * sensitivities_wrapper(xopt, *args)
    return f0val, df0dx, fval.reshape(-1, 1), dfdx


# ---------------------------------------------------------------
# Calculation with NLOPT
# ---------------------------------------------------------------


def create_nlopt_fobj(fobj, fgrad, *args):
    def fobj_nlopt(x, grad):
        f = fobj(x, *args)
        if grad.size > 0:
            grad[:] = array(fgrad(x, *args)).flatten()
        return f
    return fobj_nlopt


def create_nlopt_fconstr(fconstr, fjac, *args):
    def fconstr_nlopt(results, x, grad):
        results[:] = -1 * fconstr(x, *args).flatten()
        if grad.size > 0:
            grad[:] = -1 * fjac(x, *args)
        return
    return fconstr_nlopt



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import scriptcontext as sc

import compas_rhino

__commandname__ = "TNO_form_assign_supports"


def RunCommand(is_interactive):

    if 'TNO' not in sc.sticky:
        compas_rhino.display_message('TNO has not been initialised yet.')
        return

    scene = sc.sticky['TNO']['scene']

    objects = scene.find_by_name('Form')
    if not objects:
        compas_rhino.display_message("There is no FormDiagram in the scene.")
        return
    form = objects[0]

    form.diagram.vertices_attribute('is_fixed', False)

    show_vertices = form.settings['show.vertices']
    form.settings['show.vertices'] = True if not form.settings['show.vertices'] else show_vertices

    scene.update()

    vertices = form.select_vertices("Fix selected vertices (unfix all others)")
    if not vertices:
        return

    form.diagram.vertices_attribute('is_fixed', True, keys=vertices)
    form.settings['show.vertices'] = show_vertices

    scene.update()
    scene.save()


# ==============================================================================
# Main
# ==============================================================================
if __name__ == '__main__':

    RunCommand(True)



from numpy import array
from numpy import ones
from numpy import zeros
from numpy import concatenate
from numpy import linspace

from compas_tno.shapes import rectangular_topology
from compas_tno.shapes import MeshDos

import math


def cross_vault_highfields_proxy(xy_span=[[0.0, 10.0], [0.0, 10.0]], thk=0.5, tol=10e-6, t=0.0, discretisation=[100, 100], *args, **kwargs):
    """Function that computes the highfield of cross vaults through the proxy

    Parameters
    ----------
    xy_span : [list[float], list[float]], optional
        xy-span of the shape, by default [[0.0, 10.0], [0.0, 10.0]]
    thk : float, optional
        The thickness of the vault, by default 0.5
    tol : float, optional
        Tolerance, by default 10e-6
    t : float, optional
        Parameter for lower bound in nodes in the boundary, by default 0.0
    discretisation : list|int, optional
        Level of discretisation of the shape, by default [100, 100]

    Returns
    -------
    intradosdata
        Data to a Mesh for the intrados of the shape
    extradosdata
        Data to a Mesh for the extrados of the shape
    middledata
        Data to a Mesh for the middle of the shape
    """
    intrados, extrados, middle = cross_vault_highfields(xy_span=xy_span, thk=thk, tol=tol, t=t, discretisation=discretisation, expanded=False)
    return intrados.to_data(), extrados.to_data(), middle.to_data()


def cross_vault_highfields(xy_span=[[0.0, 10.0], [0.0, 10.0]], thk=0.50, tol=10e-6, t=0.0, discretisation=[100, 100], expanded=False):
    """ Set Cross-Vault heights.

    Parameters
    ----------
    xy_span : [list[float], list[float]], optional
        xy-span of the shape, by default [[0.0, 10.0], [0.0, 10.0]]
    thk : float, optional
        The thickness of the vault, by default 0.5
    tol : float, optional
        Tolerance, by default 10e-6
    t : float, optional
        Parameter for lower bound in nodes in the boundary, by default 0.0
    discretisation : list|int, optional
        Level of discretisation of the shape, by default [100, 100]

    Returns
    -------
    intrados : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the intrados of the shape
    extrados : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the extrados of the shape
    middle : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the middle of the shape

    Notes
    ----------------------
    Position of the quadrants is as in the schema below:

        Q3
    Q2      Q1
        Q4

    """

    if isinstance(discretisation, int):
        discretisation = [discretisation, discretisation]

    y1 = xy_span[1][1]
    y0 = xy_span[1][0]
    x1 = xy_span[0][1]
    x0 = xy_span[0][0]

    density_x = discretisation[0]
    density_y = discretisation[1]
    x = linspace(x0, x1, num=density_x+1, endpoint=True)  # arange(x0, x1 + dx/density_x, dx/density_x)
    y = linspace(y0, y1, num=density_y+1, endpoint=True)  # arange(y0, y1 + dy/density_y, dy/density_y)

    xi, yi, faces_i = rectangular_topology(x, y)

    zt = crossvault_middle_update(xi, yi, t, xy_span=xy_span, tol=1e-6)
    xyzt = array([xi, yi, zt.flatten()]).transpose()
    middle = MeshDos.from_vertices_and_faces(xyzt, faces_i)

    if expanded:
        x = concatenate(([x0 - thk/2], x, [x1 + thk/2]))
        y = concatenate(([y0 - thk/2], y, [y1 + thk/2]))
        xi, yi, faces_i = rectangular_topology(x, y)
        zub, zlb = crossvault_ub_lb_update(xi, yi, thk, t, xy_span=xy_span, tol=1e-6)
    else:
        zub, zlb = crossvault_ub_lb_update(xi, yi, thk, t, xy_span=xy_span, tol=1e-6)

    xyzub = array([xi, yi, zub.flatten()]).transpose()
    xyzlb = array([xi, yi, zlb.flatten()]).transpose()

    extrados = MeshDos.from_vertices_and_faces(xyzub, faces_i)
    intrados = MeshDos.from_vertices_and_faces(xyzlb, faces_i)

    return intrados, extrados, middle


def crossvault_ub_lb_update(x, y, thk, t, xy_span=[[0.0, 10.0], [0.0, 10.0]], tol=1e-6):
    """Update upper and lower bounds of an crossvault based in the parameters

    Parameters
    ----------
    x : list
        x-coordinates of the points
    y : list
        y-coordinates of the points
    thk : float
        Thickness of the arch
    t : float
        Parameter for lower bound in nodes in the boundary
    xy_span : [list[float], list[float]], optional
        xy-span of the shape, by default [[0.0, 10.0], [0.0, 10.0]]
    tol : float, optional
        Tolerance, by default 10e-6

    Returns
    -------
    ub : array
        Values of the upper bound in the points
    lb : array
        Values of the lower bound in the points
    """

    y1 = xy_span[1][1]
    y0 = xy_span[1][0]
    x1 = xy_span[0][1]
    x0 = xy_span[0][0]

    y1_ub = y1 + thk/2
    y0_ub = y0 - thk/2
    x1_ub = x1 + thk/2
    x0_ub = x0 - thk/2

    y1_lb = y1 - thk/2
    y0_lb = y0 + thk/2
    x1_lb = x1 - thk/2
    x0_lb = x0 + thk/2

    rx_ub = (x1_ub - x0_ub)/2
    ry_ub = (y1_ub - y0_ub)/2
    rx_lb = (x1_lb - x0_lb)/2
    ry_lb = (y1_lb - y0_lb)/2

    hc_ub = max(rx_ub, ry_ub)
    hc_lb = max(rx_lb, ry_lb)

    ub = ones((len(x), 1))
    lb = ones((len(x), 1)) * - t

    for i in range(len(x)):
        xi, yi = x[i], y[i]
        xd_ub = x0_ub + (x1_ub - x0_ub)/(y1_ub - y0_ub) * (yi - y0_ub)
        yd_ub = y0_ub + (y1_ub - y0_ub)/(x1_ub - x0_ub) * (xi - x0_ub)
        hxd_ub = math.sqrt((rx_ub)**2 - ((xd_ub - x0_ub) - rx_ub)**2)
        hyd_ub = math.sqrt((ry_ub)**2 - ((yd_ub - y0_ub) - ry_ub)**2)

        intrados_null = False

        if (yi > y1_lb and (xi > x1_lb or xi < x0_lb)) or (yi < y0_lb and (xi > x1_lb or xi < x0_lb)):
            intrados_null = True
        else:
            yi_intra = yi
            xi_intra = xi
            if yi > y1_lb:
                yi_intra = y1_lb
            if yi < y0_lb:
                yi_intra = y0_lb
            elif xi > x1_lb:
                xi_intra = x1_lb
            elif xi < x0_lb:
                xi_intra = x0_lb

            xd_lb = x0_lb + (x1_lb - x0_lb)/(y1_lb - y0_lb) * (yi_intra - y0_lb)
            yd_lb = y0_lb + (y1_lb - y0_lb)/(x1_lb - x0_lb) * (xi_intra - x0_lb)
            hxd_lb = _sqrt(((rx_lb)**2 - ((xd_lb - x0_lb) - rx_lb)**2))
            hyd_lb = _sqrt(((ry_lb)**2 - ((yd_lb - y0_lb) - ry_lb)**2))

        if yi <= y0 + (y1 - y0)/(x1 - x0) * (xi - x0) + tol and yi >= y1 - (y1 - y0)/(x1 - x0) * (xi - x0) - tol:  # Q1
            ub[i] = hc_ub*(hxd_ub + math.sqrt((ry_ub)**2 - ((yi - y0_ub) - ry_ub)**2))/(rx_ub + ry_ub)
            if not intrados_null:
                lb[i] = hc_lb*(hxd_lb + math.sqrt((ry_lb)**2 - ((yi_intra - y0_lb) - ry_lb)**2))/(rx_lb + ry_lb)
        elif yi >= y0 + (y1 - y0)/(x1 - x0) * (xi - x0) - tol and yi >= y1 - (y1 - y0)/(x1 - x0) * (xi - x0) - tol:  # Q3
            ub[i] = hc_ub*(hyd_ub + math.sqrt((rx_ub)**2 - ((xi - x0_ub) - rx_ub)**2))/(rx_ub + ry_ub)
            if not intrados_null:
                lb[i] = hc_lb*(hyd_lb + math.sqrt((rx_lb)**2 - ((xi_intra - x0_lb) - rx_lb)**2))/(rx_lb + ry_lb)
        elif yi >= y0 + (y1 - y0)/(x1 - x0) * (xi - x0) - tol and yi <= y1 - (y1 - y0)/(x1 - x0) * (xi - x0) + tol:  # Q2
            ub[i] = hc_ub*(hxd_ub + math.sqrt((ry_ub)**2 - ((yi - y0_ub) - ry_ub)**2))/(rx_ub + ry_ub)
            if not intrados_null:
                lb[i] = hc_lb*(hxd_lb + math.sqrt((ry_lb)**2 - ((yi_intra - y0_lb) - ry_lb)**2))/(rx_lb + ry_lb)
        elif yi <= y0 + (y1 - y0)/(x1 - x0) * (xi - x0) + tol and yi <= y1 - (y1 - y0)/(x1 - x0) * (xi - x0) + tol:  # Q4
            ub[i] = hc_ub*(hyd_ub + math.sqrt((rx_ub)**2 - ((xi - x0_ub) - rx_ub)**2))/(rx_ub + ry_ub)
            if not intrados_null:
                lb[i] = hc_lb*(hyd_lb + math.sqrt((rx_lb)**2 - ((xi_intra - x0_lb) - rx_lb)**2))/(rx_lb + ry_lb)
        else:
            print('Error Q. (x,y) = ({0},{1})'.format(xi, yi))

    return ub, lb


def crossvault_dub_dlb(x, y, thk, t, xy_span=[[0.0, 10.0], [0.0, 10.0]], tol=1e-6):
    """Computes the sensitivities of upper and lower bounds in the x, y coordinates and thickness specified.

    Parameters
    ----------
    x : list
        x-coordinates of the points
    y : list
        y-coordinates of the points
    thk : float
        Thickness of the arch
    t : float
        Parameter for lower bound in nodes in the boundary
    xy_span : [list[float], list[float]], optional
        xy-span of the shape, by default [[0.0, 10.0], [0.0, 10.0]]
    tol : float, optional
        Tolerance, by default 10e-6

    Returns
    -------
    dub : array
        Values of the sensitivities for the upper bound in the points
    dlb : array
        Values of the sensitivities for the lower bound in the points
    """

    y1 = xy_span[1][1]
    y0 = xy_span[1][0]
    x1 = xy_span[0][1]
    x0 = xy_span[0][0]

    y1_ub = y1 + thk/2
    y0_ub = y0 - thk/2
    x1_ub = x1 + thk/2
    x0_ub = x0 - thk/2

    y1_lb = y1 - thk/2
    y0_lb = y0 + thk/2
    x1_lb = x1 - thk/2
    x0_lb = x0 + thk/2

    rx_ub = (x1_ub - x0_ub)/2
    ry_ub = (y1_ub - y0_ub)/2
    rx_lb = (x1_lb - x0_lb)/2
    ry_lb = (y1_lb - y0_lb)/2

    hc_ub = max(rx_ub, ry_ub)
    hc_lb = max(rx_lb, ry_lb)

    ub = ones((len(x), 1))
    lb = ones((len(x), 1)) * - t
    dub = zeros((len(x), 1))  # dzub / dt
    dlb = zeros((len(x), 1))  # dzlb / dt

    dubdx = zeros((len(x), len(x)))
    dubdy = zeros((len(x), len(x)))
    dlbdx = zeros((len(x), len(x)))
    dlbdy = zeros((len(x), len(x)))

    yc = ry_ub + y0_ub  # Only works for square
    xc = rx_ub + x0_ub

    for i in range(len(x)):

        xi, yi = x[i], y[i]
        xd_ub = x0_ub + (x1_ub - x0_ub)/(y1_ub - y0_ub) * (yi - y0_ub)
        yd_ub = y0_ub + (y1_ub - y0_ub)/(x1_ub - x0_ub) * (xi - x0_ub)
        hxd_ub = math.sqrt((rx_ub)**2 - ((xd_ub - x0_ub) - rx_ub)**2)
        hyd_ub = math.sqrt((ry_ub)**2 - ((yd_ub - y0_ub) - ry_ub)**2)

        intrados_null = False

        if (yi > y1_lb and (xi > x1_lb or xi < x0_lb)) or (yi < y0_lb and (xi > x1_lb or xi < x0_lb)):
            intrados_null = True
        else:
            yi_intra = yi
            xi_intra = xi
            if yi > y1_lb:
                yi_intra = y1_lb
            elif yi < y0_lb:
                yi_intra = y0_lb
            if xi > x1_lb:
                xi_intra = x1_lb
            elif xi < x0_lb:
                xi_intra = x0_lb

            xd_lb = x0_lb + (x1_lb - x0_lb)/(y1_lb - y0_lb) * (yi_intra - y0_lb)
            yd_lb = y0_lb + (y1_lb - y0_lb)/(x1_lb - x0_lb) * (xi_intra - x0_lb)
            hxd_lb = _sqrt(((rx_lb)**2 - ((xd_lb - x0_lb) - rx_lb)**2))
            hyd_lb = _sqrt(((ry_lb)**2 - ((yd_lb - y0_lb) - ry_lb)**2))

        if yi <= y0 + (y1 - y0)/(x1 - x0) * (xi - x0) + tol and yi >= y1 - (y1 - y0)/(x1 - x0) * (xi - x0) - tol:  # Q1
            ub[i] = hc_ub*(hxd_ub + math.sqrt((ry_ub)**2 - ((yi - y0_ub) - ry_ub)**2))/(rx_ub + ry_ub)
            dub[i] = 1/2 * ry_ub/ub[i] * hc_ub/((rx_ub + ry_ub)/2)
            # dubdx[i, i] += 0.0
            dubdy[i, i] += - (yi - yc) / ub[i]
            if not intrados_null:
                lb[i] = hc_lb*(hxd_lb + math.sqrt((ry_lb)**2 - ((yi_intra - y0_lb) - ry_lb)**2))/(rx_lb + ry_lb)
                dlb[i] = - 1/2 * ry_lb/lb[i] * hc_lb/((rx_lb + ry_lb)/2)
                # dlbdx[i, i] += 0.0
                dlbdy[i, i] += - (yi - yc) / lb[i]
        if yi >= y0 + (y1 - y0)/(x1 - x0) * (xi - x0) - tol and yi >= y1 - (y1 - y0)/(x1 - x0) * (xi - x0) - tol:  # Q3
            ub[i] = hc_ub*(hyd_ub + math.sqrt((rx_ub)**2 - ((xi - x0_ub) - rx_ub)**2))/(rx_ub + ry_ub)
            dub[i] = 1/2 * rx_ub/ub[i] * hc_ub/((rx_ub + ry_ub)/2)
            # dubdy[i, i] += 0.0
            dubdx[i, i] += - (xi - xc) / ub[i]
            if not intrados_null:
                lb[i] = hc_lb*(hyd_lb + math.sqrt((rx_lb)**2 - ((xi_intra - x0_lb) - rx_lb)**2))/(rx_lb + ry_lb)
                dlb[i] = - 1/2 * rx_lb/lb[i] * hc_lb/((rx_lb + ry_lb)/2)
                # dlbdy[i, i] += 0.0
                dlbdx[i, i] += - (xi - xc) / lb[i]
        if yi >= y0 + (y1 - y0)/(x1 - x0) * (xi - x0) - tol and yi <= y1 - (y1 - y0)/(x1 - x0) * (xi - x0) + tol:  # Q2
            ub[i] = hc_ub*(hxd_ub + math.sqrt((ry_ub)**2 - ((yi - y0_ub) - ry_ub)**2))/(rx_ub + ry_ub)
            dub[i] = 1/2 * ry_ub/ub[i] * hc_ub/((rx_ub + ry_ub)/2)
            # dubdx[i, i] += 0.0
            dubdy[i, i] += - (yi - yc) / ub[i]
            if not intrados_null:
                lb[i] = hc_lb*(hxd_lb + math.sqrt((ry_lb)**2 - ((yi_intra - y0_lb) - ry_lb)**2))/(rx_lb + ry_lb)
                dlb[i] = - 1/2 * ry_lb/lb[i] * hc_lb/((rx_lb + ry_lb)/2)
                # dlbdx[i, i] += 0.0
                dlbdy[i, i] += - (yi - yc) / lb[i]
        if yi <= y0 + (y1 - y0)/(x1 - x0) * (xi - x0) + tol and yi <= y1 - (y1 - y0)/(x1 - x0) * (xi - x0) + tol:  # Q4
            ub[i] = hc_ub*(hyd_ub + math.sqrt((rx_ub)**2 - ((xi - x0_ub) - rx_ub)**2))/(rx_ub + ry_ub)
            dub[i] = 1/2 * rx_ub/ub[i] * hc_ub/((rx_ub + ry_ub)/2)
            # dubdy[i, i] += 0.0
            dubdx[i, i] += - (xi - xc) / ub[i]
            if not intrados_null:
                lb[i] = hc_lb*(hyd_lb + math.sqrt((rx_lb)**2 - ((xi_intra - x0_lb) - rx_lb)**2))/(rx_lb + ry_lb)
                dlb[i] = - 1/2 * rx_lb/lb[i] * hc_lb/((rx_lb + ry_lb)/2)
                # dlbdy[i, i] += 0.0
                dlbdx[i, i] += - (xi - xc) / lb[i]
        # else:
        #     print('Error Q. (x,y) = ({0},{1})'.format(xi, yi))

    return dub, dlb, dubdx, dubdy, dlbdx, dlbdy  # ub, lb


def crossvault_middle_update(x, y, t, xy_span=[[0.0, 10.0], [0.0, 10.0]], tol=1e-6):
    """Update middle of a crossvault based in the parameters

    Parameters
    ----------
    x : list
        x-coordinates of the points
    y : list
        y-coordinates of the points
    thk : float
        Thickness of the arch
    t : float
        Parameter for lower bound in nodes in the boundary
    xy_span : [list[float], list[float]], optional
        xy-span of the shape, by default [[0.0, 10.0], [0.0, 10.0]]
    tol : float, optional
        Tolerance, by default 10e-6

    Returns
    -------
    z : array
        Values of the middle surface in the points
    """

    y1 = xy_span[1][1]
    y0 = xy_span[1][0]
    x1 = xy_span[0][1]
    x0 = xy_span[0][0]

    rx = (x1 - x0)/2
    ry = (y1 - y0)/2
    hc = max(rx, ry)

    z = zeros((len(x), 1))

    for i in range(len(x)):
        xi, yi = x[i], y[i]
        if yi > y1:
            yi = y1
        if yi < y0:
            yi = y0
        if xi > x1:
            xi = x1
        if xi < x0:
            xi = x0
        xd = x0 + (x1 - x0)/(y1 - y0) * (yi - y0)
        yd = y0 + (y1 - y0)/(x1 - x0) * (xi - x0)
        hxd = math.sqrt(abs((rx)**2 - ((xd - x0) - rx)**2))
        hyd = math.sqrt(abs((ry)**2 - ((yd - y0) - ry)**2))
        if yi <= y0 + (y1 - y0)/(x1 - x0) * (xi - x0) + tol and yi >= y1 - (y1 - y0)/(x1 - x0) * (xi - x0) - tol:  # Q1
            z[i] = hc*(hxd + math.sqrt((ry)**2 - ((yi - y0) - ry)**2))/(rx + ry)
        elif yi >= y0 + (y1 - y0)/(x1 - x0) * (xi - x0) - tol and yi >= y1 - (y1 - y0)/(x1 - x0) * (xi - x0) - tol:  # Q3
            z[i] = hc*(hyd + math.sqrt((rx)**2 - ((xi - x0) - rx)**2))/(rx + ry)
        elif yi >= y0 + (y1 - y0)/(x1 - x0) * (xi - x0) - tol and yi <= y1 - (y1 - y0)/(x1 - x0) * (xi - x0) + tol:  # Q2
            z[i] = hc*(hxd + math.sqrt((ry)**2 - ((yi - y0) - ry)**2))/(rx + ry)
        elif yi <= y0 + (y1 - y0)/(x1 - x0) * (xi - x0) + tol and yi <= y1 - (y1 - y0)/(x1 - x0) * (xi - x0) + tol:  # Q4
            z[i] = hc*(hyd + math.sqrt((rx)**2 - ((xi - x0) - rx)**2))/(rx + ry)
        else:
            print('Vertex did not belong to any Q. (x,y) = ({0},{1})'.format(xi, yi))
            z[i] = -t

    return z


def _sqrt(x):
    try:
        sqrt_x = math.sqrt(x)
    except BaseException:
        if x > -10e4:
            sqrt_x = math.sqrt(abs(x))
        else:
            sqrt_x = 0.0
            print('Problems to sqrt: ', x)
    return sqrt_x


# def crossvault_dub_dlb_old(x, y, thk, t, xy_span=[[0.0, 10.0], [0.0, 10.0]], tol=1e-6):

#     y1 = xy_span[1][1]
#     y0 = xy_span[1][0]
#     x1 = xy_span[0][1]
#     x0 = xy_span[0][0]

#     y1_ub = y1 + thk/2
#     y0_ub = y0 - thk/2
#     x1_ub = x1 + thk/2
#     x0_ub = x0 - thk/2

#     y1_lb = y1 - thk/2
#     y0_lb = y0 + thk/2
#     x1_lb = x1 - thk/2
#     x0_lb = x0 + thk/2

#     rx_ub = (x1_ub - x0_ub)/2
#     ry_ub = (y1_ub - y0_ub)/2
#     rx_lb = (x1_lb - x0_lb)/2
#     ry_lb = (y1_lb - y0_lb)/2

#     hc_ub = max(rx_ub, ry_ub)
#     hc_lb = max(rx_lb, ry_lb)

#     ub = ones((len(x), 1))
#     lb = ones((len(x), 1)) * - t
#     dub = zeros((len(x), 1))
#     dlb = zeros((len(x), 1))

#     for i in range(len(x)):

#         xi, yi = x[i], y[i]
#         xd_ub = x0_ub + (x1_ub - x0_ub)/(y1_ub - y0_ub) * (yi - y0_ub)
#         yd_ub = y0_ub + (y1_ub - y0_ub)/(x1_ub - x0_ub) * (xi - x0_ub)
#         hxd_ub = math.sqrt((rx_ub)**2 - ((xd_ub - x0_ub) - rx_ub)**2)
#         hyd_ub = math.sqrt((ry_ub)**2 - ((yd_ub - y0_ub) - ry_ub)**2)

#         intrados_null = False

#         if (yi > y1_lb and (xi > x1_lb or xi < x0_lb)) or (yi < y0_lb and (xi > x1_lb or xi < x0_lb)):
#             intrados_null = True
#         else:
#             yi_intra = yi
#             xi_intra = xi
#             if yi > y1_lb:
#                 yi_intra = y1_lb
#             elif yi < y0_lb:
#                 yi_intra = y0_lb
#             if xi > x1_lb:
#                 xi_intra = x1_lb
#             elif xi < x0_lb:
#                 xi_intra = x0_lb

#             xd_lb = x0_lb + (x1_lb - x0_lb)/(y1_lb - y0_lb) * (yi_intra - y0_lb)
#             yd_lb = y0_lb + (y1_lb - y0_lb)/(x1_lb - x0_lb) * (xi_intra - x0_lb)
#             hxd_lb = _sqrt(((rx_lb)**2 - ((xd_lb - x0_lb) - rx_lb)**2))
#             hyd_lb = _sqrt(((ry_lb)**2 - ((yd_lb - y0_lb) - ry_lb)**2))

#         if yi <= y0 + (y1 - y0)/(x1 - x0) * (xi - x0) + tol and yi >= y1 - (y1 - y0)/(x1 - x0) * (xi - x0) - tol:  # Q1
#             ub[i] = hc_ub*(hxd_ub + math.sqrt((ry_ub)**2 - ((yi - y0_ub) - ry_ub)**2))/(rx_ub + ry_ub)
#             dub[i] = 1/2 * ry_ub/ub[i] * hc_ub/((rx_ub + ry_ub)/2)
#             if not intrados_null:
#                 lb[i] = hc_lb*(hxd_lb + math.sqrt((ry_lb)**2 - ((yi_intra - y0_lb) - ry_lb)**2))/(rx_lb + ry_lb)
#                 dlb[i] = - 1/2 * ry_lb/lb[i] * hc_lb/((rx_lb + ry_lb)/2)
#         elif yi >= y0 + (y1 - y0)/(x1 - x0) * (xi - x0) - tol and yi >= y1 - (y1 - y0)/(x1 - x0) * (xi - x0) - tol:  # Q3
#             ub[i] = hc_ub*(hyd_ub + math.sqrt((rx_ub)**2 - ((xi - x0_ub) - rx_ub)**2))/(rx_ub + ry_ub)
#             dub[i] = 1/2 * rx_ub/ub[i] * hc_ub/((rx_ub + ry_ub)/2)
#             if not intrados_null:
#                 lb[i] = hc_lb*(hyd_lb + math.sqrt((rx_lb)**2 - ((xi_intra - x0_lb) - rx_lb)**2))/(rx_lb + ry_lb)
#                 dlb[i] = - 1/2 * rx_lb/lb[i] * hc_lb/((rx_lb + ry_lb)/2)
#         elif yi >= y0 + (y1 - y0)/(x1 - x0) * (xi - x0) - tol and yi <= y1 - (y1 - y0)/(x1 - x0) * (xi - x0) + tol:  # Q2
#             ub[i] = hc_ub*(hxd_ub + math.sqrt((ry_ub)**2 - ((yi - y0_ub) - ry_ub)**2))/(rx_ub + ry_ub)
#             dub[i] = 1/2 * ry_ub/ub[i] * hc_ub/((rx_ub + ry_ub)/2)
#             if not intrados_null:
#                 lb[i] = hc_lb*(hxd_lb + math.sqrt((ry_lb)**2 - ((yi_intra - y0_lb) - ry_lb)**2))/(rx_lb + ry_lb)
#                 dlb[i] = - 1/2 * ry_lb/lb[i] * hc_lb/((rx_lb + ry_lb)/2)
#         elif yi <= y0 + (y1 - y0)/(x1 - x0) * (xi - x0) + tol and yi <= y1 - (y1 - y0)/(x1 - x0) * (xi - x0) + tol:  # Q4
#             ub[i] = hc_ub*(hyd_ub + math.sqrt((rx_ub)**2 - ((xi - x0_ub) - rx_ub)**2))/(rx_ub + ry_ub)
#             dub[i] = 1/2 * rx_ub/ub[i] * hc_ub/((rx_ub + ry_ub)/2)
#             if not intrados_null:
#                 lb[i] = hc_lb*(hyd_lb + math.sqrt((rx_lb)**2 - ((xi_intra - x0_lb) - rx_lb)**2))/(rx_lb + ry_lb)
#                 dlb[i] = - 1/2 * rx_lb/lb[i] * hc_lb/((rx_lb + ry_lb)/2)
#         else:
#             print('Error Q. (x,y) = ({0},{1})'.format(xi, yi))

#     return dub, dlb  # ub, lb


# def crossvault_ub_lb_update_repetitive(x, y, thk, t, xy_span=[[0.0, 10.0], [0.0, 10.0]], tol=1e-6):  # Invoke the method crossvault middle many times

#     ub = ones((len(x), 1))
#     lb = ones((len(x), 1)) * - t

#     rx = (xy_span[0][1] - xy_span[0][0])/2
#     ry = (xy_span[1][1] - xy_span[1][0])/2

#     if ry > rx:
#         y1_ub = xy_span[1][1] + thk/2 * rx/ry
#         y0_ub = xy_span[1][0] - thk/2 * rx/ry
#         x1_ub = xy_span[0][1] + thk/2
#         x0_ub = xy_span[0][0] - thk/2
#         xy_span_ub = [[x0_ub, x1_ub], [y0_ub, y1_ub]]

#         y1_lb = xy_span[1][1] - thk/2 * rx/ry
#         y0_lb = xy_span[1][0] + thk/2 * rx/ry
#         x1_lb = xy_span[0][1] - thk/2
#         x0_lb = xy_span[0][0] + thk/2
#         xy_span_lb = [[x0_lb, x1_lb], [y0_lb, y1_lb]]
#     else:
#         y1_ub = xy_span[1][1] + thk/2
#         y0_ub = xy_span[1][0] - thk/2
#         x1_ub = xy_span[0][1] + thk/2 * ry/rx
#         x0_ub = xy_span[0][0] - thk/2 * ry/rx
#         xy_span_ub = [[x0_ub, x1_ub], [y0_ub, y1_ub]]

#         y1_lb = xy_span[1][1] - thk/2
#         y0_lb = xy_span[1][0] + thk/2
#         x1_lb = xy_span[0][1] - thk/2 * ry/rx
#         x0_lb = xy_span[0][0] + thk/2 * ry/rx
#         xy_span_lb = [[x0_lb, x1_lb], [y0_lb, y1_lb]]

#     ub = crossvault_middle_update(x, y, t, xy_span=xy_span_ub)
#     lb = crossvault_middle_update(x, y, t, xy_span=xy_span_lb)

#     return ub, lb
